Package: spBPS
Title: Bayesian Predictive Stacking for Scalable Geospatial Transfer
        Learning
Version: 2.0-1
Authors@R: c(
    person("Luca", "Presicce", role = c("aut", "cre"), email = "l.presicce@campus.unimib.it", comment = c(ORCID = "0009-0005-7062-3523")),
    person("Sudipto", "Banerjee", role = "aut"))
Maintainer: Luca Presicce <l.presicce@campus.unimib.it>
Author: Luca Presicce [aut, cre] (ORCID:
    <https://orcid.org/0009-0005-7062-3523>),
  Sudipto Banerjee [aut]
Description: Provides functions for Bayesian Predictive Stacking within the Bayesian transfer learning framework for geospatial artificial systems, as introduced in "Bayesian Transfer Learning for Artificially Intelligent Geospatial Systems: A Predictive Stacking Approach" (Presicce and Banerjee, 2025) <doi:10.48550/arXiv.2410.09504>. This methodology enables efficient Bayesian geostatistical modeling, utilizing predictive stacking to improve inference across spatial datasets. The core functions leverage 'C++' for high-performance computation, making the framework well-suited for large-scale spatial data analysis in parallel and distributed computing environments. Designed for scalability, it allows seamless application in computationally demanding scenarios.
Depends: R (>= 1.8.0)
Imports: Rcpp, CVXR (>= 1.8.1), mniw
LinkingTo: Rcpp, RcppArmadillo
Suggests: knitr, rmarkdown, abind, mvnfast, ECOSolveR, foreach,
        parallel, doParallel, tictoc, MBA, RColorBrewer, classInt, sp,
        fields, testthat (>= 3.0.0)
Config/testthat/edition: 3
License: GPL (>= 3)
Encoding: UTF-8
RoxygenNote: 7.3.3
VignetteBuilder: knitr
URL: https://lucapresicce.github.io/spBPS/
NeedsCompilation: yes
Packaged: 2026-05-14 14:03:06 UTC; presi
Repository: CRAN
Date/Publication: 2026-05-14 14:50:02 UTC
Built: R 4.6.0; x86_64-apple-darwin20; 2026-05-14 15:43:08 UTC; unix
Archs: spBPS.so.dSYM

# Generated by roxygen2: do not edit by hand

S3method(predict,spBPS)
S3method(print,spBPS)
S3method(print,spBPS_pred)
S3method(print,spBPS_pred_summary)
export(bayesMvLMconjugate)
export(pred_bayesMvLMconjugate)
export(spBPS)
export(subset_data)
importFrom(CVXR,Maximize)
importFrom(CVXR,Problem)
importFrom(CVXR,Variable)
importFrom(CVXR,psolve)
importFrom(CVXR,status)
importFrom(CVXR,value)
importFrom(Rcpp,sourceCpp)
importFrom(mniw,rMNorm)
importFrom(mniw,riwish)
importFrom(utils,setTxtProgressBar)
importFrom(utils,txtProgressBar)
useDynLib(spBPS, .registration = TRUE)

# spBPS <a href="https://lucapresicce.github.io/spBPS/"><img src="man/figures/logo.png" align="right" height="138" alt="spBPS website" /></a>

<!-- badges: end -->
[![CRAN_Status_Badge](https://www.r-pkg.org/badges/version/spBPS?color=blue)](https://CRAN.R-project.org/package=spBPS)
[![Downloads](https://cranlogs.r-pkg.org/badges/spBPS?color=orange)](https://CRAN.R-project.org/package=spBPS)
[![CRAN RStudio mirror downloads](https://cranlogs.r-pkg.org/badges/grand-total/spBPS?color=green)](https://CRAN.R-project.org/package=spBPS)


<!--
[![R-CMD-check](https://github.com/lucapresicce/spBPS/actions/workflows/Standard-CI.yaml/badge.svg)](https://github.com/lucapresicce/spBPS/actions/workflows/Standard-CI.yaml)

# Univariate and Multivariate Accelerated Spatial Modeling by Bayesian Predictive Stacking
-->

## Overview

This package provides the principal functions to perform accelerated modeling for univariate and multivariate spatial regressions. The package is used mostly within the novel working paper *"Bayesian Transfer Learning for Artificially Intelligent Geospatial Systems: A Predictive Stacking Approach" ([**Luca Presicce**](https://lucapresicce.github.io/) and Sudipto Banerjee, 2024+)"*. To guarantee the reproducibility of scientific results, in the [Bayesian-Transfer-Learning-for-GeoAI](https://github.com/lucapresicce/Bayesian-Transfer-Learning-for-GeoAI) repository are also available all the scripts of code used for simulations, data analysis, and results presented in the Manuscript and its Supplemental material.


## Installation

If installing from CRAN, use the following.

```r
install.packages("spBPS")
```

For a quick installation of the development version, run the following command 
in R. We use the `devtools` R package to install. Then, check for its presence
on your device, otherwise install it:

```r
if (!require(devtools)) {
  install.packages("devtools", dependencies = TRUE)
}
```

Once you have installed *devtools*, we can proceed. Let's install the `spBPS` package!

```r
devtools::install_github("lucapresicce/spBPS")
```


## Usage
Once successfully installed, load the library in R.

```r
library(spBPS)
```
Cool! You are ready to start, now you too could perform **_fast & feasible_** Bayesian geostatistical modeling!

<!--
## Tutorial for usage
-->


## Contacts

| | |
| :--- | :---: |
| Author | Luca Presicce (l.presicce@campus.unimib.it) & Sudipto Banerjee (sudipto@ucla.edu) |
| Maintainer | Luca Presicce (l.presicce@campus.unimib.it) |
| Reference | [**Luca Presicce**](https://lucapresicce.github.io/) and Sudipto Banerjee (2024+) *"Bayesian Transfer Learning for Artificially Intelligent Geospatial Systems: A Predictive Stacking Approach"*  |

<!--
Maintainer: l.presicce@campus.unimib.it
Reference: **Luca Presicce** and Sudipto Banerjee (2024+) *"Accelerated Meta-Kriging for massive Spatial dataset"* 
-->

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----setup--------------------------------------------------------------------
library(spBPS)

## ----warning=FALSE, message=F-------------------------------------------------
library(foreach)
library(parallel)
library(doParallel)
library(tictoc)
library(MBA)
library(classInt)
library(RColorBrewer)
library(sp)
library(fields)
library(mvnfast)
library(abind)

## ----results=F----------------------------------------------------------------
# dimensions
n <- 1000
u <- 250
p <- 2
q <- 1

# parameters
B <- c(-0.75, 1.85)
tau2 <- 0.25
sigma2 <- 1
delta <- tau2/sigma2
phi <- 4

set.seed(4-8-15-16-23-42)
# generate sintethic data
crd <- matrix(runif((n+u) * 2), ncol = 2)
X_or <- cbind(rep(1, n+u), matrix(runif((p-1)*(n+u)), ncol = (p-1)))
D <- spBPS:::arma_dist(crd)
Rphi <- exp(-phi * D)
W_or <- matrix(0, n+u) + mniw::rmNorm(1, rep(0, n+u), sigma2*Rphi)
Y_or <- X_or %*% B + W_or + mniw::rmNorm(1, rep(0, n+u), diag(delta*sigma2, n+u))

# train data
crd_s <- crd[1:n, ]
X <- X_or[1:n, ]
W <- W_or[1:n, ]
Y <- Y_or[1:n, ]

# prediction data
crd_u <- crd[-(1:n), ]
X_u <- X_or[-(1:n), ]
W_u <- W_or[-(1:n), ]
Y_u <- Y_or[-(1:n), ]

## ----results=F----------------------------------------------------------------
# priors 
priors <- list(mu_B = matrix(0, nrow = p, ncol = q),
               V_r = diag(10, p),
               Psi = diag(1, q),
               nu = 3)

# hyperparameters values
alfa_seq <- c(0.7, 0.8, 0.9)
phi_seq <- c(3, 4, 5)
hyperpar <- list(alpha = alfa_seq, phi = phi_seq)

## ----results=F----------------------------------------------------------------
# subset dimension
subset_size <- 500

# number of posterior draws
R <- 200

# number of computational cores
n_core <- 1

## ----results=F----------------------------------------------------------------
out <- spBPS(data = list(Y = Y, X = X),
      priors = priors,
      coords = crd_s,
      hyperpar = hyperpar,
      subset_size = subset_size,
      combine_method = "bps",
      draws = R,
      newdata = list(X = X_u, coords = crd_u),
      cores = n_core)

## -----------------------------------------------------------------------------
# statistics computations W
pred_mat_W <- do.call(abind, c(lapply(out$predictive, function(x) x$Wu), along = 3))
post_mean_W <- apply(pred_mat_W, c(1,2), mean)
post_qnt_W <- apply(pred_mat_W, c(1,2), quantile, c(0.025, 0.975))

# Empirical coverage for W
coverage_W <- mean(W_u >= post_qnt_W[1,,1] & W_u <= post_qnt_W[2,,1])
cat("Empirical coverage for Spatial process:", round(coverage_W, 3),"\n")

# statistics computations Y
pred_mat_Y <- do.call(abind, c(lapply(out$predictive, function(x) x$Yu), along = 3))
post_mean_Y <- apply(pred_mat_Y, c(1,2), mean)
post_qnt_Y <- apply(pred_mat_Y, c(1,2), quantile, c(0.025, 0.975))

# Empirical coverage for Y
coverage_Y <- mean(Y_u >= post_qnt_Y[1,,1] & Y_u <= post_qnt_Y[2,,1])
cat("Empirical coverage for Response:", round(coverage_Y, 3),"\n")

# Root Mean Square Prediction Error
rmspe_W <- sqrt( mean( (W_u - post_mean_W)^2 ) )
rmspe_Y <- sqrt( mean( (Y_u - post_mean_Y)^2 ) )
cat("RMSPE for Spatial process:", round(rmspe_W, 3), "\n")
cat("RMSPE for Response:", round(rmspe_Y, 3), "\n")

## ----results=F, echo=F--------------------------------------------------------
# True spatial process surface interpolation
h <- 12
surf.W <- MBA::mba.surf(cbind(crd_s, W), no.X = 500, no.Y = 500,
                        exten = TRUE, sp = TRUE, h = h)$xyz.est
surf.brks <- classIntervals(surf.W$z, 100, 'pretty')$brks
col.pal <- colorRampPalette(brewer.pal(11,'RdBu')[11:1])
xlim <- c(0, 1)
zlim <- range(surf.W$z)

# image for plot
iw <- as.image.SpatialGridDataFrame(surf.W)

# BPS surfaces interpolation
h <- 12
surf.Wp <- MBA::mba.surf(cbind(crd_u, post_mean_W), no.X = 500, no.Y = 500,
                         exten = TRUE, sp = TRUE, h = h)$xyz.est
zlimp <- range(surf.Wp$z)

# image for plot
iwp <- as.image.SpatialGridDataFrame(surf.Wp)

## ----echo=F, fig.dim = c(7.25, 4)---------------------------------------------
# Plotting
oldpar <- par(no.readonly = TRUE)
par(mfrow = c(1, 2))

plot(crd, type="n", cex=0.5, xlim=xlim, axes=FALSE, ylab="Northing", xlab="Easting",
     main="Spatial process")
axis(2, las=1)
axis(1)
image.plot(iw, add=TRUE, col=rev(col.pal(length(surf.brks)-1)), zlim=zlim)

plot(crd, type="n", cex=0.5, xlim=xlim, axes=F, ylab="Northing", xlab="Easting")
title(main="Spatial process (Prediction)")
mtext(side = 3, paste("RMSPE :", round(rmspe_W, 3)))
axis(2, las=1)
axis(1)
image.plot(iwp, add=TRUE, col=rev(col.pal(length(surf.brks)-1)), zlim=zlimp)

par(oldpar)

## ----results=F, echo=F--------------------------------------------------------
# True response surface interpolation
h <- 12
surf.Y <- MBA::mba.surf(cbind(crd_s, Y), no.X = 500, no.Y = 500,
                        exten = TRUE, sp = TRUE, h = h)$xyz.est
surf.brks <- classIntervals(surf.Y$z, 100, 'pretty')$brks
col.pal <- colorRampPalette(brewer.pal(11,'RdBu')[11:1])
xlim <- c(0, 1)
zlim <- range(surf.Y$z)

# image for plot
iy <- as.image.SpatialGridDataFrame(surf.Y)

# BPS surfaces interpolation
h <- 12
surf.Yp <- MBA::mba.surf(cbind(crd_u, post_mean_Y), no.X = 500, no.Y = 500,
                         exten = TRUE, sp = TRUE, h = h)$xyz.est
zlimp <- range(surf.Yp$z)

# image for plot
iyp <- as.image.SpatialGridDataFrame(surf.Yp)

## ----echo=F, fig.dim = c(7.25, 4)---------------------------------------------
# Plotting
oldpar <- par(no.readonly = TRUE)
par(mfrow = c(1, 2))

plot(crd, type="n", cex=0.5, xlim=xlim, axes=FALSE, ylab="Northing", xlab="Easting",
     main="Response")
axis(2, las=1)
axis(1)
image.plot(iy, add=TRUE, col=rev(col.pal(length(surf.brks)-1)), zlim=zlim)

plot(crd, type="n", cex=0.5, xlim=xlim, axes=F, ylab="Northing", xlab="Easting")
title(main="Response (Prediction)")
mtext(side = 3, paste("RMSPE :", round(rmspe_Y, 3)))
axis(2, las=1)
axis(1)
image.plot(iyp, add=TRUE, col=rev(col.pal(length(surf.brks)-1)), zlim=zlimp)

par(oldpar)

# packages
# library(spBPS)
# devtools::install(upgrade = "never", build_vignettes = FALSE, dependencies = FALSE)
library(mniw)
library(tictoc)
library(parallel)
library(doParallel)
library(foreach)

# dimensions
set.seed(2025)
n <- 1000
u <- 100
p <- 2
q <- 2

# parameters
B <- matrix(c(-0.75, 0.90, 1.85, -1.1), p, q)
sigma2 <- matrix(c(1, 0.5, 0.5, 1), q, q)
# B <- matrix(c(-0.75, -1.1), p, q)
# sigma2 <- 1
alfa <- 0.8
phi <- 4

set.seed(97)
# generate sintethic data
crd <- matrix(runif((n+u) * 2), ncol = 2)
X_or <- cbind(rep(1, n+u), matrix(runif((p-1)*(n+u)), ncol = (p-1)))
D <- spBPS:::arma_dist(crd)
gc()
Rphi <- exp(-phi * D)
rm("D"); gc()
W_or <- matrix(0, n+u, q) + mniw::rMNorm(1, Lambda = matrix(0, n+u, q), SigmaR = Rphi, SigmaC = sigma2)
rm("Rphi"); gc()
Y_or <- X_or %*% B + W_or + mniw::rMNorm(1, Lambda = matrix(0, n+u, q), SigmaR = diag((1/alfa)-1, n+u), SigmaC = sigma2)
gc()

# sample data
crd_s <- crd[1:n, ]
X <- X_or[1:n, ]
W <- W_or[1:n, ]
Y <- matrix(Y_or[1:n, ], n, q)

# prediction data
crd_u <- crd[-(1:n), ]
X_u <- X_or[-(1:n), ]
W_u <- W_or[-(1:n), ]
Y_u <- matrix(Y_or[-(1:n), ], u, q)

# # hyperparameters values
# alfa_seq <- c(0.7, 0.8, 0.9)
# phi_seq <- c(3, 4, 5)
#
# # function for the fit loop
# fit_loop <- function(i) {
#
#   Yi <- data_part$Y_list[[i]]; Xi <- data_part$X_list[[i]]; crd_i <- data_part$crd_list[[i]]
#   p <- ncol(Xi); q <- ncol(Yi)
#   bps <- spBPS:::BPS_weights_MvT(data = list(Y = Yi, X = Xi),
#                                priors = list(mu_B = matrix(0, nrow = p, ncol = q),
#                                              V_r = diag(10, p),
#                                              Psi = diag(1, q),
#                                              nu = 3), coords = crd_i,
#                                hyperpar = list(alpha = alfa_seq, phi = phi_seq), K = 5)
#   w_hat <- bps$W
#   epd <- bps$epd
#
#   result <- list(epd, w_hat)
#   return(result)
#
# }
#
# # function for the pred loop
# pred_loop <- function(r) {
#
#   ind_s <- subset_ind[1]
#   Ys <- data_part$Y_list[[ind_s]]; Xs <- data_part$X_list[[ind_s]]; crds <- data_part$crd_list[[ind_s]]; Ws <- W_list[[ind_s]]
#   result <- spBPS:::BPS_post_MvT(data = list(Y = Ys, X = Xs), coords = crds,
#                                X_u = X_u, crd_u = crd_u,
#                                priors = list(mu_B = matrix(0, nrow = p, ncol = q),
#                                              V_r = diag(10, p),
#                                              Psi = diag(1, q),
#                                              nu = 3),
#                                hyperpar = list(alpha = alfa_seq, phi = phi_seq),
#                                W = Ws, R = 1)
#
#   return(result)
# }
#
# # subsetting data
# subset_size <- 200
# K <- n/subset_size
# data_part <- spBPS::subset_data(data = list(Y = Y, X = X, crd = crd_s), K = K)
#
# # timing
# tic("total")
#
# # number of clusters for parallel implementation
# n.core <- parallel::detectCores(logical=F)-1
#
# # list of function
# funs_fit <- lsf.str()[which(lsf.str() != "fit_loop")]
#
# # list of function
# funs_pred <- lsf.str()[which(lsf.str() != "pred_loop")]
#
# # starting cluster
# cl <- makeCluster(n.core)
# registerDoParallel(cl)
#
# # parallelized subset computation of GP in different cores
# tic("fit")
# obj_fit <- foreach(i = 1:K, .noexport = funs_fit) %dopar% { fit_loop(i) }
# fit_time <- toc()
#
# gc()
# # Combination using double BPS
# tic("comb")
# comb_bps <- spBPS:::BPS_combine(obj_fit, K, 1)
# comb_time <- toc()
# Wbps <- comb_bps$W
# W_list <- comb_bps$W_list
#
# gc()
# # parallelized subset computation of GP in different cores
# R <- 200
# subset_ind <- sample(1:K, R, T, Wbps)
# tic("prediction")
# predictions <- foreach(r = 1:R, .noexport = funs_pred) %dopar% { pred_loop(r) }
# prd_time <- toc()
#
# # closing cluster
# stopCluster(cl)
# gc()
#
# # timing
# tot_time <- toc()


# New unified function: spBPS()

# priors
priors <- list(mu_B = matrix(0, nrow = p, ncol = q),
               V_r = diag(10, p),
               Psi = diag(1, q),
               nu = 3)

# hyperparameters values
alfa_seq <- c(0.7, 0.8, 0.9)
phi_seq <- c(3, 4, 5)
hyperpar <- list(alpha = alfa_seq, phi = phi_seq)

tic("spBPS")
out <- spBPS::spBPS(data = list(Y = Y, X = X),
# out <- spBPS(data = list(Y = Y, X = X),
      priors = priors,
      coords = crd_s,
      hyperpar = hyperpar,
      subset_size = 200,
      cv_folds = 5,
      rp = 1,
      combine_method = "bps",
      draws = 200,
      newdata = list(X = X_u, coords = crd_u),
      cores = parallel::detectCores(logical=F)-1)
new_time <- toc()

# Compare results

# Compare times
tot_time$callback_msg
new_time$callback_msg

# Compare predictions
pred_matrix_legacy <- do.call(abind::abind, c(lapply(predictions, function(x) x$Pred[[1]]$Yu), along = 3))
pred_matrix_orch <- do.call(abind::abind, c(lapply(out$predictive, function(x) x$Yu), along = 3))

# Adjusting for disagreement
disagree_array <- do.call(abind::abind, c(lapply(predictions, function(x) x$Pred[[1]]$MY), along = 3))
pred_matrix_legacy <- pred_matrix_legacy - disagree_array
agreement <- apply(disagree_array, c(1,2), mean)
agree_array <- replicate(R, agreement, simplify = "array")
pred_matrix_legacy <- pred_matrix_legacy + agree_array

# Check equality
all.equal(pred_matrix_legacy, pred_matrix_orch, tolerance = 1e-5)

# Check root mean squared prediciton error (RMSPE) and coverage
post_mean_legacy <- apply(pred_matrix_legacy, c(1,2), mean)
post_qnt_legacy <- apply(pred_matrix_legacy, c(1,2), quantile, c(0.025, 0.975))
c(mean(Y_u[,1] >= post_qnt_legacy[1,,1] & Y_u[,1] <= post_qnt_legacy[2,,1]),
                mean(Y_u[,2] >= post_qnt_legacy[1,,2] & Y_u[,2] <= post_qnt_legacy[2,,2]))
mean(post_qnt_legacy[2,,]-post_qnt_legacy[1,,])
rmspe_legacy <- sqrt( colMeans( (Y_u - post_mean_legacy)^2 ) )
mean(rmspe_legacy)

post_mean_orch <- apply(pred_matrix_orch, c(1,2), mean)
post_qnt_orch <- apply(pred_matrix_orch, c(1,2), quantile, c(0.025, 0.975))
c(mean(Y_u[,1] >= post_qnt_orch[1,,1] & Y_u[,1] <= post_qnt_orch[2,,1]),
                mean(Y_u[,2] >= post_qnt_orch[1,,2] & Y_u[,2] <= post_qnt_orch[2,,2]))
mean(post_qnt_orch[2,,]-post_qnt_orch[1,,])
rmspe_orch <- sqrt( colMeans( (Y_u - post_mean_orch)^2 ) )
mean(rmspe_orch)

# Compare weights
Wbps
out$weights_global

# Compare posterior samples
beta_smp <- sapply(1:R, function(r){predictions[[r]]$Post[[1]]$beta[1:p,]}, simplify = "array")
(post_mean_beta <-  apply(beta_smp, c(1,2), mean))
post_low_beta <- apply(beta_smp, c(1,2), quantile, c(0.025))
post_upp_beta <- apply(beta_smp, c(1,2), quantile, c(0.975))

sigma_smp <- sapply(1:R, function(r){predictions[[r]]$Post[[1]]$sigma}, simplify = "array")
(post_mean_sigma <- apply(sigma_smp, c(1,2), mean))
post_low_sigma <- apply(sigma_smp, c(1,2), quantile, c(0.025))
post_upp_sigma <- apply(sigma_smp, c(1,2), quantile, c(0.975))

(post_mean_hyp <- sapply(1:K, function(k)t(spBPS:::expand_grid_cpp(hyperpar$alpha, hyperpar$phi)) %*% W_list[[k]]) %*% Wbps)
(post_var_hyp <- sapply(1:K, function(k)t(spBPS:::expand_grid_cpp(hyperpar$alpha, hyperpar$phi)) %*% W_list[[k]])^2 %*% Wbps - (post_mean_hyp^2))

# collecting
posterior_bps <- cbind(t(sapply(1:R, function(r)matrix(beta_smp[,,r]))),
                        t(sapply(1:R, function(r)matrix(sigma_smp[,,r])))[,-3],
                        cbind(rep(post_mean_hyp[2], R), rep(post_mean_hyp[2], R)))
colnames(posterior_bps) <- c("beta[1,1]", "beta[2,1]", "beta[1,2]", "beta[2,2]", "Sigma[1,1]", "Sigma[2,1]", "Sigma[2,2]", "phi[1]", "phi[2]")

# fixing true parameter for plotting
true_par <- c(matrix(B), matrix(sigma2)[-3], rep(phi, 2))

# plotting bps
library(ggplot2)
library(bayesplot)
bayesplot_theme_set(theme_default(base_size = 14, base_family = "sans"))
color_scheme_set("viridis")
post_int_bps <- mcmc_recover_intervals(posterior_bps, true_par,
                                        prob = 0.95,
                                        prob_outer = 0.95,
                                        point_est = "mean",
                                        size = 4,
                                        alpha = 0.75) +
  scale_x_discrete(labels = c(expression(beta[list(1,1)]), expression(beta[list(2,1)]), expression(beta[list(3,1)]), expression(beta[list(1,2)]), expression(beta[list(2,2)]), expression(beta[list(3,2)]),
                              expression(phi[1]), expression(phi[2]),
                              expression(Sigma[list(1,1)]), expression(Sigma[list(2,1)]), expression(Sigma[list(2,2)]))) +
  ggtitle("Credible intervals BPS",
          "with posterior means, true values, and 95% credible intervals")
post_int_bps


beta_smp <- sapply(1:R, function(r){out$posterior[[r]]$beta[1:p,]}, simplify = "array")
(post_mean_beta <-  apply(beta_smp, c(1,2), mean))
post_low_beta <- apply(beta_smp, c(1,2), quantile, c(0.025))
post_upp_beta <- apply(beta_smp, c(1,2), quantile, c(0.975))

sigma_smp <- sapply(1:R, function(r){out$posterior[[r]]$sigma}, simplify = "array")
(post_mean_sigma <- apply(sigma_smp, c(1,2), mean))
post_low_sigma <- apply(sigma_smp, c(1,2), quantile, c(0.025))
post_upp_sigma <- apply(sigma_smp, c(1,2), quantile, c(0.975))

(post_mean_hyp <- sapply(1:K, function(k)t(spBPS:::expand_grid_cpp(hyperpar$alpha, hyperpar$phi)) %*% matrix(out$weights_local[[k]])) %*% matrix(out$weights_global))
(post_var_hyp <- sapply(1:K, function(k)t(spBPS:::expand_grid_cpp(hyperpar$alpha, hyperpar$phi)) %*% matrix(out$weights_local[[k]]))^2 %*% matrix(out$weights_global) - (post_mean_hyp^2))

# collecting
posterior_bps <- cbind(t(sapply(1:R, function(r)matrix(beta_smp[,,r]))),
                        t(sapply(1:R, function(r)matrix(sigma_smp[,,r])))[,-3],
                        cbind(rep(post_mean_hyp[2], R), rep(post_mean_hyp[2], R)))
colnames(posterior_bps) <- c("beta[1,1]", "beta[2,1]", "beta[1,2]", "beta[2,2]", "Sigma[1,1]", "Sigma[2,1]", "Sigma[2,2]", "phi[1]", "phi[2]")

# fixing true parameter for plotting
true_par <- c(matrix(B), matrix(sigma2)[-3], rep(phi, 2))

# plotting bps
library(ggplot2)
library(bayesplot)
bayesplot_theme_set(theme_default(base_size = 14, base_family = "sans"))
color_scheme_set("viridis")
post_int_bps2 <- mcmc_recover_intervals(posterior_bps, true_par,
                                        prob = 0.95,
                                        prob_outer = 0.95,
                                        point_est = "mean",
                                        size = 4,
                                        alpha = 0.75) +
  scale_x_discrete(labels = c(expression(beta[list(1,1)]), expression(beta[list(2,1)]), expression(beta[list(3,1)]), expression(beta[list(1,2)]), expression(beta[list(2,2)]), expression(beta[list(3,2)]),
                              expression(phi[1]), expression(phi[2]),
                              expression(Sigma[list(1,1)]), expression(Sigma[list(2,1)]), expression(Sigma[list(2,2)]))) +
  ggtitle("Credible intervals BPS",
          "with posterior means, true values, and 95% credible intervals")
post_int_bps2

#' Gibbs sampler for Conjugate Bayesian Multivariate Linear Models
#'
#' @param Y [matrix] \eqn{n \times q} of response variables
#' @param X [matrix] \eqn{n \times p} of predictors
#' @param mu_B [matrix] \eqn{p \times q} prior mean for \eqn{\beta}
#' @param V_B [matrix] \eqn{p \times p} prior row covariance for \eqn{\beta}
#' @param nu [double] prior parameter for \eqn{\Sigma}
#' @param Psi [matrix] prior parameter for \eqn{\Sigma}
#' @param n_iter [integer] iteration number for Gibbs sampler
#' @param burn_in [integer] number of burn-in iteration
#'
#' @return B_samples [array] of posterior sample for \eqn{\beta}
#' @return Sigma_samples [array] of posterior samples for \eqn{\Sigma}
#'
#' @importFrom mniw rMNorm riwish
#' @importFrom utils txtProgressBar setTxtProgressBar
#'
#' @examples
#' \donttest{
#' ## Generate data
#' n <- 100
#' p <- 3
#' q <- 2
#' Y <- matrix(rnorm(n*q), nrow = n, ncol = q)
#' X <- matrix(rnorm(n*p), nrow = n, ncol = p)
#'
#' ## Prior parameters
#' mu_B <- matrix(0, p, q)
#' V_B <- diag(10, p)
#' nu <- 3
#' Psi <- diag(q)
#'
#' ## Samples from posteriors
#' n_iter <- 1000
#' burn_in <- 500
#' set.seed(1234)
#' samples <- spBPS::bayesMvLMconjugate(Y, X, mu_B, V_B, nu, Psi, n_iter, burn_in)
#' }
#'
#' @export
bayesMvLMconjugate <- function(Y, X, mu_B, V_B, nu, Psi, n_iter = 1000, burn_in = 500) {
  n <- nrow(Y)
  q <- ncol(Y)
  p <- ncol(X)

  # Initialize matrices to store samples
  B_samples <- array(0, dim = c(p, q, n_iter))
  Sigma_samples <- array(0, dim = c(q, q, n_iter))

  # Initial values
  B <- matrix(0, p, q)
  Sigma <- diag(q)

  pb <- txtProgressBar(0, n_iter, style = 3)

  # Gibbs sampler iterations
  for (iter in 1:n_iter) {
    # Sample B given Sigma and data
    V_n <- solve(solve(V_B) + t(X) %*% X)
    B_n <- V_n %*% (solve(V_B) %*% mu_B + t(X) %*% Y)
    B <- mniw::rMNorm(1, B_n, V_n, Sigma)

    # Sample Sigma given B and data
    S_n <- Psi + t(Y - X %*% B) %*% (Y - X %*% B)
    Sigma <- mniw::riwish(1, nu = nu + n, Psi = S_n)

    # Store samples
    B_samples[, , iter] <- B
    Sigma_samples[, , iter] <- Sigma

    setTxtProgressBar(pb, iter)
  }

  # Remove burn-in samples
  B_samples <- B_samples[, , (burn_in + 1):n_iter]
  Sigma_samples <- Sigma_samples[, , (burn_in + 1):n_iter]

  list(B_samples = B_samples, Sigma_samples = Sigma_samples)
}

#' Predictive sampler for Conjugate Bayesian Multivariate Linear Models
#'
#' @param X_new [matrix] \eqn{n_new \times p} of predictors for new data points
#' @param B_samples [array] of posterior sample for \eqn{\beta}
#' @param Sigma_samples [array] of posterior samples for \eqn{\Sigma}
#'
#' @return Y_pred [matrix] of posterior mean for response matrix Y predictions
#' @return Y_pred_samples [array] of posterior predictive sample for response matrix Y
#'
#' @importFrom mniw rMNorm
#' @importFrom utils txtProgressBar setTxtProgressBar
#'
#' @examples
#' \donttest{
#' ## Generate data
#' n <- 100
#' p <- 3
#' q <- 2
#' Y <- matrix(rnorm(n*q), nrow = n, ncol = q)
#' X <- matrix(rnorm(n*p), nrow = n, ncol = p)
#'
#' ## Prior parameters
#' mu_B <- matrix(0, p, q)
#' V_B <- diag(10, p)
#' nu <- 3
#' Psi <- diag(q)
#'
#' ## Samples from posteriors
#' n_iter <- 1000
#' burn_in <- 500
#' set.seed(1234)
#' samples <- spBPS::bayesMvLMconjugate(Y, X, mu_B, V_B, nu, Psi, n_iter, burn_in)
#'
#' ## Extract posterior samples
#' B_samples <- samples$B_samples
#' Sigma_samples <- samples$Sigma_samples
#'
#' ## Samples from predictive posterior (based posterior samples)
#' m <- 50
#' X_new <- matrix(rnorm(m*p), nrow = m, ncol = p)
#' pred <- spBPS::pred_bayesMvLMconjugate(X_new, B_samples, Sigma_samples)
#' }
#'
#' @export
pred_bayesMvLMconjugate <- function(X_new, B_samples, Sigma_samples) {
  n_iter <- dim(B_samples)[3]
  n_new <- nrow(X_new)
  q <- dim(Sigma_samples)[1]

  # Initialize matrix to store predictions
  Y_pred_samples <- array(0, dim = c(n_new, q, n_iter))
  Y_pred <- matrix(0, n_new, q)

  # Loop over the posterior samples
  pb <- txtProgressBar(0, n_iter, style = 3)
  for (iter in 1:n_iter) {
    B <- B_samples[, , iter]
    Sigma <- Sigma_samples[, , iter]

    # Predict responses for the new data
    Y_pred_iter <- mniw::rMNorm(n = 1, Lambda = X_new %*% B, SigmaR = diag(n_new), SigmaC = Sigma)

    # Store samples
    Y_pred_samples[, , iter] <- Y_pred_iter

    # Add the predictions to the cumulative sum
    Y_pred <- Y_pred + Y_pred_iter

    setTxtProgressBar(pb, iter)
  }

  # Average the predictions over all posterior samples
  Y_pred <- Y_pred / n_iter

  list(Y_pred_samples = Y_pred_samples, Y_pred = Y_pred)

}

#' @keywords internal
"_PACKAGE"

## usethis namespace: start
#' @useDynLib spBPS, .registration = TRUE
## usethis namespace: end
NULL

## usethis namespace: start
#' @importFrom Rcpp sourceCpp
## usethis namespace: end
NULL

#' Unified spatial BPS workflow (multivariate path, works for q = 1)
#'
#' @description
#' Orchestrates subsetting, local stacking weight estimation, global stacking combination, and optional posterior or predictive simulation using Double Bayesian Predictive Stacking for latent spatial regression. Works for both multivariate outcomes and the univariate case via `q = 1`.
#'
#' @param data List with matrices `Y` (response, n x q) and `X` (covariates, n x p).
#' @param priors List of priors for the multivariate model: `mu_B` (p x q mean
#'   matrix), `V_r` (p x p covariance), `Psi` (q x q scale), `nu` (degrees of
#'   freedom).
#' @param coords Matrix of observation coordinates (n x d).
#' @param hyperpar List with elements `alpha` and `phi` (vectors allowed); defines
#'   the grid of models over which stacking weights are computed.
#' @param subset_size Target subset size when `K` is not provided. Default 500.
#' @param K Optional number of subsets. When `NULL`, computed as
#'   `ceiling(nrow(Y) / subset_size)` and lower-bounded at 1.
#' @param cv_folds Number of folds for local cross-validation. Default 5.
#' @param rp Fraction of rows used when recomputing global stacking weights
#'   (passed to `BPS_combine`). Ignored when `combine_method = "pseudoBMA"`.
#' @param combine_method Global combination method: Bayesian Predictive Stacking
#'   (`"bps"`) or pseudo-BMA (`"pseudoBMA"`).
#' @param draws Number of posterior/predictive draws to return (0 to skip
#'   sampling). When positive and `newdata` is supplied, joint posterior and
#'   predictive draws are returned. When positive and `newdata` is `NULL`,
#'   only posterior draws (beta, sigma) are returned.
#' @param newdata Optional list with `X` (u x p) and `coords` (u x d) for
#'   prediction locations. Required when `draws > 0` and predictions are desired.
#'   For large `u`, use `pred_batch_size` to control memory usage.
#' @param include_latent Unused; kept for compatibility.
#' @param n_cores Number of cores for parallel computation. Controls both the
#'   local weight estimation (parallel over K subsets) and predictive sampling.
#'   Default 1.
#' @param pred_batch_size Batch size for streaming prediction. Controls the
#'   maximum number of prediction sites processed at once, hence the peak memory
#'   usage. Default 200. Rule of thumb: peak RAM (MB) ? batch_size x q x draws x 8 / 1e6.
#'   Set `NULL` for automatic selection (min(200, u)).
#'
#' @return Object of class `"spBPS"` ? a list with components:
#'   \describe{
#'     \item{subsets}{Partition information: Y_list, X_list, crd_list.}
#'     \item{weights_global}{K-vector of global stacking weights.}
#'     \item{weights_local}{K-list of local stacking weight vectors.}
#'     \item{epd}{K-list of n_k x J log-density matrices.}
#'     \item{priors, hyperpar}{Stored for use by \code{\link{predict.spBPS}}.}
#'     \item{timings}{Named numeric vector: fitting, combination, sampling (seconds).}
#'     \item{posterior}{(if draws > 0) List with three-dimensional arrays:
#'       \code{beta} (p x q x R), \code{sigma} (q x q x R), \code{model} (R).}
#'     \item{predictive}{(if draws > 0 and newdata supplied) List with
#'       three-dimensional arrays: \code{Wu}, \code{Yu}, \code{MY} (each u x q x R).}
#'   }
#'
#' @seealso \code{\link{predict.spBPS}} for generating predictions on new data
#'   after fitting.
#'
#' @examples
#' \donttest{
#' n <- 1000
#' p <- 2
#' q <- 1
#'
#' Y <- matrix(rnorm(n * q), ncol = q)
#' X <- matrix(rnorm(n * p), ncol = p)
#' coords <- matrix(runif(n * 2), ncol = 2)
#'
#' data    <- list(Y = Y, X = X)
#' priors  <- list(mu_B = matrix(0, nrow = p, ncol = q),
#'                 V_r  = diag(10, p),
#'                 Psi  = diag(1, q),
#'                 nu   = 3)
#' hyperpar <- list(alpha = 0.5, phi = 1)
#'
#' res <- spBPS(data, priors, coords, hyperpar, subset_size = 200)
#' }
#'
#' @export
spBPS <- function(data,
                  priors,
                  coords,
                  hyperpar,
                  subset_size     = 500L,
                  K               = NULL,
                  cv_folds        = 5L,
                  rp              = 1.0,
                  combine_method  = c("bps", "pseudoBMA"),
                  draws           = 0L,
                  newdata         = NULL,
                  include_latent  = FALSE,
                  n_cores         = 1L,
                  pred_batch_size = 200L) {

  cat("\n====================================================\n")
  cat("         Welcome to spBPS Bayesian Engine\n")
  cat("====================================================\n\n")

  combine_method <- match.arg(combine_method)
  timings <- c(fitting = NA_real_, combination = NA_real_, sampling = NA_real_)

  # -- Normalise inputs ------------------------------------------------------
  data$Y <- as.matrix(data$Y)
  data$X <- as.matrix(data$X)
  coords  <- as.matrix(coords)
  n <- nrow(data$Y)

  if (is.null(K))
    K <- max(1L, min(n, as.integer(ceiling(n / subset_size))))

  Grid <- expand_grid_cpp(hyperpar$alpha, hyperpar$phi)
  J    <- nrow(Grid)

  # -- Step 1: Partition -----------------------------------------------------
  cat("Partitioning data into K =", K, "subsets ...\n\n")
  subsets <- subset_data(list(Y = data$Y, X = data$X, crd = coords), K = K)

  # -- Step 2: Local weight estimation --------------------------------------
  cat("Computing local stacking weights over J =", J, "models ...\n")
  t_fit <- system.time({
    fit_list <- compute_all_local_weights_v2(
      Y_list   = subsets$Y_list,
      X_list   = subsets$X_list,
      crd_list = subsets$crd_list,
      priors   = priors,
      hyperpar = hyperpar,
      cv_folds = cv_folds,
      n_cores  = n_cores)
  })
  timings["fitting"] <- t_fit["elapsed"]
  cat(sprintf("Local weights computed in %.1f s.\n\n", t_fit["elapsed"]))

  # -- Step 3: Global combination --------------------------------------------
  cat("Computing global stacking weights over K =", K, "partitions ...\n")
  t_comb <- system.time({
    if (combine_method == "bps") {
      comb <- BPS_combine_v2(fit_list = fit_list, K = K, rp = rp)
    } else {
      comb <- BPS_PseudoBMA_v2(fit_list = fit_list, n_cores = n_cores)
    }
  })
  timings["combination"] <- t_comb["elapsed"]
  W_global <- as.numeric(comb$W)
  W_local  <- comb$W_list
  cat(sprintf("Global weights computed in %.1f s.\n\n", t_comb["elapsed"]))

  out <- list(
    subsets        = subsets,
    weights_global = W_global,
    weights_local  = W_local,
    epd            = lapply(fit_list, `[[`, 1L),
    priors         = priors,
    hyperpar       = hyperpar,
    timings        = timings
  )
  class(out) <- "spBPS"

  # -- Step 4: Posterior / predictive sampling -------------------------------
  if (draws > 0L) {

    K_loc      <- length(subsets$Y_list)
    W_loc_vecs <- lapply(W_local, as.numeric)
    subset_ind <- sample.int(K_loc, draws, replace = TRUE, prob = W_global) - 1L

    if (is.null(newdata)) {
      # Posterior-only draws
      cat("Posterior sampling (R =", draws, "draws) ...\n")
      t_samp <- system.time({
        res_post <- BPS_post_all_subsets_v2(
          Y_list       = subsets$Y_list,
          X_list       = subsets$X_list,
          crd_list     = subsets$crd_list,
          W_local_list = W_loc_vecs,
          subset_ind   = as.integer(subset_ind),
          priors       = priors,
          hyperpar     = hyperpar,
          n_cores      = n_cores)
      })
      timings["sampling"] <- t_samp["elapsed"]
      out$posterior <- list(beta  = res_post$beta,
                            sigma = res_post$sigma,
                            model = res_post$model)
      cat(sprintf("Posterior sampling completed in %.1f s.\n\n", t_samp["elapsed"]))

    } else {
      # Posterior + predictive draws
      newdata$X      <- as.matrix(newdata$X)
      newdata$coords <- as.matrix(newdata$coords)
      X_u   <- newdata$X
      crd_u <- newdata$coords
      u     <- nrow(X_u)

      B_eff <- if (is.null(pred_batch_size) || pred_batch_size <= 0L)
                 min(200L, u)
               else
                 min(as.integer(pred_batch_size), u)

      cat(sprintf("Posterior & Predictive Sampling (R = %d draws) ...\n", draws))
      t_samp <- system.time({
        res_all <- BPS_pred_streaming_v2(
          Y_list          = subsets$Y_list,
          X_list          = subsets$X_list,
          crd_list        = subsets$crd_list,
          W_local_list    = W_loc_vecs,
          subset_ind      = as.integer(subset_ind),
          crd_u           = crd_u,
          X_u             = X_u,
          priors          = priors,
          hyperpar        = hyperpar,
          pred_batch_size = B_eff,
          n_cores         = as.integer(n_cores),
          return_summary  = FALSE,
          prob_lo         = 0.025,
          prob_hi         = 0.975)

        res_post <- BPS_post_all_subsets_v2(
          Y_list       = subsets$Y_list,
          X_list       = subsets$X_list,
          crd_list     = subsets$crd_list,
          W_local_list = W_loc_vecs,
          subset_ind   = as.integer(subset_ind),
          priors       = priors,
          hyperpar     = hyperpar,
          n_cores      = n_cores)
      })
      timings["sampling"] <- t_samp["elapsed"]

      out$posterior  <- list(beta  = res_post$beta,
                             sigma = res_post$sigma,
                             model = res_post$model)
      out$predictive <- list(Wu = res_all$Wu,
                             Yu = res_all$Yu,
                             MY = res_all$MY)
      cat(sprintf("Posterior & Predictive sampling completed in %.1f s.\n\n",
                  t_samp["elapsed"]))
    }
  }

  out$timings <- timings

  cat("====================================================\n")
  cat("     spBPS pipeline completed successfully!\n")
  cat("====================================================\n")
  cat(sprintf("  Fitting:     %.1f s\n", timings["fitting"]))
  cat(sprintf("  Combination: %.1f s\n", timings["combination"]))
  if (!is.na(timings["sampling"]))
    cat(sprintf("  Sampling:    %.1f s\n", timings["sampling"]))
  cat(sprintf("  TOTAL:       %.1f s\n\n",
              sum(timings, na.rm = TRUE)))
  return(out)
}

#' @keywords internal
#' @export
print.spBPS <- function(x, ...) {
  cat("spBPS object\n")
  cat(sprintf("  K subsets:   %d\n", length(x$subsets$Y_list)))
  cat(sprintf("  J models:    %d\n", length(x$weights_global)))
  if (!is.null(x$posterior))
    cat(sprintf("  Posterior draws:  %d\n", dim(x$posterior$beta)[3L]))
  if (!is.null(x$predictive))
    cat(sprintf("  Predictive draws: %d  (u = %d)\n",
                dim(x$predictive$Yu)[3L], dim(x$predictive$Yu)[1L]))
  invisible(x)
}

#' @keywords internal
#' @export
print.spBPS_pred <- function(x, ...) {
  cat(sprintf("spBPS_pred  [u=%d  q=%d  R=%d  %.1f s]\n",
              x$u, dim(x$Yu)[2L], x$draws, x$elapsed))
  invisible(x)
}

#' @keywords internal
#' @export
print.spBPS_pred_summary <- function(x, ...) {
  cat(sprintf("spBPS_pred_summary  [u=%d  q=%d  R=%d  probs=(%.3f,%.3f)  %.1f s]\n",
              x$u, ncol(x$mu_Yu), x$draws,
              x$probs[1L], x$probs[2L], x$elapsed))
  invisible(x)
}

#' Predict at new locations using a fitted spBPS model
#'
#' Generates posterior predictive samples at new spatial locations using a
#' fitted \code{spBPS} object. Memory usage is controlled via
#' \code{pred_batch_size}
#'
#' @param object Object of class \code{"spBPS"} returned by \code{spBPS()}.
#' @param newdata List with elements \code{X} (u x p covariate matrix) and \code{coords} (u x d coordinate matrix).
#' @param draws Integer. Number of posterior predictive draws. Default 200.
#' @param pred_batch_size Integer. Number of prediction sites processed per batch. Default 200. Smaller values use less RAM; larger values may be faster.
#' @param n_cores Integer. Number of OMP threads. Default 1.
#' @param return_summary Logical. If \code{FALSE} (default) return full draw arrays. If \code{TRUE} return summary statistics.
#' @param probs Numeric vector of length 2. Quantile probabilities for credible intervals when \code{return_summary = TRUE}. Default \code{c(0.025, 0.975)}.
#' @param ... Currently unused.
#'
#' @return When \code{return_summary = FALSE}, a list with arrays \code{Wu},
#'   \code{Yu}, \code{MY} of dimension u x q x R.
#'   When \code{return_summary = TRUE}, a list with u x q matrices
#'   \code{mu_Yu}, \code{sd_Yu}, \code{q_lo}, \code{q_hi}, \code{mu_Wu}.
#'
#' @seealso \code{\link{spBPS}}
#'
#' @examples
#' \donttest{
#' n <- 1000
#' p <- 2
#' q <- 1
#'
#' Y <- matrix(rnorm(n * q), ncol = q)
#' X <- matrix(rnorm(n * p), ncol = p)
#' coords <- matrix(runif(n * 2), ncol = 2)
#'
#' data    <- list(Y = Y, X = X)
#' priors  <- list(mu_B = matrix(0, nrow = p, ncol = q),
#'                 V_r  = diag(10, p),
#'                 Psi  = diag(1, q),
#'                 nu   = 3)
#' hyperpar <- list(alpha = 0.5, phi = 1)
#'
#' res <- spBPS(data, priors, coords, hyperpar, subset_size = 200)
#'
#' u <- 500
#' p <- 2
#' q <- 1
#'
#' X_u <- matrix(rnorm(u * p), ncol = p)
#' crd_u <- matrix(runif(u * 2), ncol = 2)
#'
#' # Predictive sampling
#' pred <- predict(res, newdata=list(X=X_u, coords=crd_u),
#'                 draws=200L, pred_batch_size=500L)
#'
#' # Summary statistics only (memory-efficient for large u)
#' pred_sum <- predict(res, newdata=list(X=X_u, coords=crd_u),
#'                     draws=500L, return_summary=TRUE,
#'                     probs=c(0.025, 0.975))
#' }
#'
#'
#' @method predict spBPS
#' @export
predict.spBPS <- function(object,
                          newdata,
                          draws           = 200L,
                          pred_batch_size = 200L,
                          n_cores         = 1L,
                          return_summary  = FALSE,
                          probs           = c(0.025, 0.975),
                          ...) {

  if (!inherits(object, "spBPS"))
    stop("object must be of class 'spBPS' (output of spBPS()).")
  if (is.null(object$subsets) || is.null(object$priors) || is.null(object$hyperpar))
    stop("object is missing $subsets, $priors, or $hyperpar. Re-run spBPS().")
  if (!is.list(newdata) || is.null(newdata$X) || is.null(newdata$coords))
    stop("newdata must be a list with elements $X and $coords.")
  if (length(probs) != 2L || probs[1L] >= probs[2L] ||
      any(probs < 0) || any(probs > 1))
    stop("probs must be a length-2 vector with 0 <= probs[1] < probs[2] <= 1.")

  newdata$X      <- as.matrix(newdata$X)
  newdata$coords <- as.matrix(newdata$coords)
  X_u   <- newdata$X
  crd_u <- newdata$coords
  u     <- nrow(X_u)
  q     <- ncol(object$subsets$Y_list[[1L]])
  K     <- length(object$subsets$Y_list)

  B_eff <- if (is.null(pred_batch_size) || pred_batch_size <= 0L)
              min(200L, u) else min(as.integer(pred_batch_size), u)

  W_global   <- object$weights_global
  W_loc_vecs <- lapply(object$weights_local, as.numeric)
  subset_ind <- sample.int(K, draws, replace = TRUE, prob = W_global) - 1L

  cat(sprintf("[predict.spBPS]  u = %d  draws = %d  cores = %d\n",
              u, draws, n_cores))

  t_pred <- system.time({
    res <- BPS_pred_streaming_v2(
      Y_list          = object$subsets$Y_list,
      X_list          = object$subsets$X_list,
      crd_list        = object$subsets$crd_list,
      W_local_list    = W_loc_vecs,
      subset_ind      = as.integer(subset_ind),
      crd_u           = crd_u,
      X_u             = X_u,
      priors          = object$priors,
      hyperpar        = object$hyperpar,
      pred_batch_size = B_eff,
      n_cores         = as.integer(n_cores),
      return_summary  = return_summary,
      prob_lo         = probs[1L],
      prob_hi         = probs[2L])
  })

  cat(sprintf("  Completed in %.1f s.\n", t_pred["elapsed"]))

  if (!return_summary) {
    structure(list(Wu = res$Wu, Yu = res$Yu, MY = res$MY,
                   u = u, draws = draws, batch_size = B_eff,
                   elapsed = t_pred["elapsed"]),
              class = "spBPS_pred")
  } else {
    structure(list(mu_Yu = res$mu_Yu, sd_Yu = res$sd_Yu,
                   q_lo  = res$q_lo,  q_hi  = res$q_hi,
                   mu_Wu = res$mu_Wu, probs = probs,
                   u = u, draws = draws, batch_size = B_eff,
                   elapsed = t_pred["elapsed"]),
              class = "spBPS_pred_summary")
  }
}

#' Solver for Bayesian Predictive Stacking of Predictive densities convex optimization problem
#'
#' @param scores [matrix] \eqn{N \times K} of expected predictive density evaluations for the K models considered
#'
#' @return W [matrix] of Bayesian Predictive Stacking weights for the K models considered
#'
#' @importFrom CVXR Variable Maximize Problem psolve status value
#'
conv_opt <- function(scores) {
  # library(CVXR, quietly = T)

  # set up minimization problem and solve it
  weights <- Variable( ncol(scores) )
  constraints <- list(weights >= 0, sum(weights) == 1)

  # the constraint for sum up to 1 with positive weights
  f <- Maximize( mean( log( scores %*% weights ) ) )
  problem <- Problem(f, constraints)

  # set the solver
  solver_to_use <- if (requireNamespace("ECOSolveR", quietly = TRUE)) {
    "ECOS_BB"
  } else {
    "SCS"  # CLARABEL, ECOS, OSQP
  }
  result <- psolve(problem, solver = solver_to_use)

  # return the weights
  W <- if(status( problem ) == "solver_error") {
    matrix(rep(1/ncol(scores), ncol(scores)))
  } else {
    value( weights )
  }
  return(W)
}

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----setup--------------------------------------------------------------------
library(spBPS)

## ----warning=FALSE, message=F-------------------------------------------------
library(foreach)
library(parallel)
library(doParallel)
library(tictoc)
library(MBA)
library(classInt)
library(RColorBrewer)
library(sp)
library(fields)
library(mvnfast)
library(abind)

## ----results=F----------------------------------------------------------------
# dimensions
n <- 1000
u <- 250
p <- 2
q <- 1

# parameters
B <- c(-0.75, 1.85)
tau2 <- 0.25
sigma2 <- 1
delta <- tau2/sigma2
phi <- 4

set.seed(4-8-15-16-23-42)
# generate sintethic data
crd <- matrix(runif((n+u) * 2), ncol = 2)
X_or <- cbind(rep(1, n+u), matrix(runif((p-1)*(n+u)), ncol = (p-1)))
D <- spBPS:::arma_dist(crd)
Rphi <- exp(-phi * D)
W_or <- matrix(0, n+u) + mniw::rmNorm(1, rep(0, n+u), sigma2*Rphi)
Y_or <- X_or %*% B + W_or + mniw::rmNorm(1, rep(0, n+u), diag(delta*sigma2, n+u))

# train data
crd_s <- crd[1:n, ]
X <- X_or[1:n, ]
W <- W_or[1:n, ]
Y <- Y_or[1:n, ]

# prediction data
crd_u <- crd[-(1:n), ]
X_u <- X_or[-(1:n), ]
W_u <- W_or[-(1:n), ]
Y_u <- Y_or[-(1:n), ]

## ----results=F----------------------------------------------------------------
# priors 
priors <- list(mu_B = matrix(0, nrow = p, ncol = q),
               V_r = diag(10, p),
               Psi = diag(1, q),
               nu = 3)

# hyperparameters values
alfa_seq <- c(0.7, 0.8, 0.9)
phi_seq <- c(3, 4, 5)
hyperpar <- list(alpha = alfa_seq, phi = phi_seq)

## ----results=F----------------------------------------------------------------
# subset dimension
subset_size <- 100

# number of posterior draws
R <- 200

# number of computational cores
n_core <- 1

## ----results=F----------------------------------------------------------------
out <- spBPS(data = list(Y = Y, X = X),
      priors = priors,
      coords = crd_s,
      hyperpar = hyperpar,
      subset_size = subset_size,
      combine_method = "bps",
      draws = R,
      newdata = list(X = X_u, coords = crd_u),
      n_cores = n_core)

## -----------------------------------------------------------------------------
# statistics computations W
pred_mat_W <- out$predictive$Wu
post_mean_W <- apply(pred_mat_W, c(1,2), mean)
post_qnt_W <- apply(pred_mat_W, c(1,2), quantile, c(0.025, 0.975))

# Empirical coverage for W
coverage_W <- mean(W_u >= post_qnt_W[1,,1] & W_u <= post_qnt_W[2,,1])
cat("Empirical coverage for Spatial process:", round(coverage_W, 3),"\n")

# statistics computations Y
pred_mat_Y <- out$predictive$Yu
post_mean_Y <- apply(pred_mat_Y, c(1,2), mean)
post_qnt_Y <- apply(pred_mat_Y, c(1,2), quantile, c(0.025, 0.975))

# Empirical coverage for Y
coverage_Y <- mean(Y_u >= post_qnt_Y[1,,1] & Y_u <= post_qnt_Y[2,,1])
cat("Empirical coverage for Response:", round(coverage_Y, 3),"\n")

# Root Mean Square Prediction Error
rmspe_W <- sqrt( mean( (W_u - post_mean_W)^2 ) )
rmspe_Y <- sqrt( mean( (Y_u - post_mean_Y)^2 ) )
cat("RMSPE for Spatial process:", round(rmspe_W, 3), "\n")
cat("RMSPE for Response:", round(rmspe_Y, 3), "\n")

## ----results=F, echo=F--------------------------------------------------------
# True spatial process surface interpolation
h <- 12
surf.W <- MBA::mba.surf(cbind(crd_s, W), no.X = 500, no.Y = 500,
                        exten = TRUE, sp = TRUE, h = h)$xyz.est
surf.brks <- classIntervals(surf.W$z, 100, 'pretty')$brks
col.pal <- colorRampPalette(brewer.pal(11,'RdBu')[11:1])
xlim <- c(0, 1)
zlim <- range(surf.W$z)

# image for plot
iw <- as.image.SpatialGridDataFrame(surf.W)

# BPS surfaces interpolation
h <- 12
surf.Wp <- MBA::mba.surf(cbind(crd_u, post_mean_W), no.X = 500, no.Y = 500,
                         exten = TRUE, sp = TRUE, h = h)$xyz.est
zlimp <- range(surf.Wp$z)

# image for plot
iwp <- as.image.SpatialGridDataFrame(surf.Wp)

## ----echo=F, fig.dim = c(7.25, 4)---------------------------------------------
# Plotting
oldpar <- par(no.readonly = TRUE)
par(mfrow = c(1, 2))

plot(crd, type="n", cex=0.5, xlim=xlim, axes=FALSE, ylab="Northing", xlab="Easting",
     main="Spatial process")
axis(2, las=1)
axis(1)
image.plot(iw, add=TRUE, col=rev(col.pal(length(surf.brks)-1)), zlim=zlim)

plot(crd, type="n", cex=0.5, xlim=xlim, axes=F, ylab="Northing", xlab="Easting")
title(main="Spatial process (Prediction)")
mtext(side = 3, paste("RMSPE :", round(rmspe_W, 3)))
axis(2, las=1)
axis(1)
image.plot(iwp, add=TRUE, col=rev(col.pal(length(surf.brks)-1)), zlim=zlimp)

par(oldpar)

## ----results=F, echo=F--------------------------------------------------------
# True response surface interpolation
h <- 12
surf.Y <- MBA::mba.surf(cbind(crd_s, Y), no.X = 500, no.Y = 500,
                        exten = TRUE, sp = TRUE, h = h)$xyz.est
surf.brks <- classIntervals(surf.Y$z, 100, 'pretty')$brks
col.pal <- colorRampPalette(brewer.pal(11,'RdBu')[11:1])
xlim <- c(0, 1)
zlim <- range(surf.Y$z)

# image for plot
iy <- as.image.SpatialGridDataFrame(surf.Y)

# BPS surfaces interpolation
h <- 12
surf.Yp <- MBA::mba.surf(cbind(crd_u, post_mean_Y), no.X = 500, no.Y = 500,
                         exten = TRUE, sp = TRUE, h = h)$xyz.est
zlimp <- range(surf.Yp$z)

# image for plot
iyp <- as.image.SpatialGridDataFrame(surf.Yp)

## ----echo=F, fig.dim = c(7.25, 4)---------------------------------------------
# Plotting
oldpar <- par(no.readonly = TRUE)
par(mfrow = c(1, 2))

plot(crd, type="n", cex=0.5, xlim=xlim, axes=FALSE, ylab="Northing", xlab="Easting",
     main="Response")
axis(2, las=1)
axis(1)
image.plot(iy, add=TRUE, col=rev(col.pal(length(surf.brks)-1)), zlim=zlim)

plot(crd, type="n", cex=0.5, xlim=xlim, axes=F, ylab="Northing", xlab="Easting")
title(main="Response (Prediction)")
mtext(side = 3, paste("RMSPE :", round(rmspe_Y, 3)))
axis(2, las=1)
axis(1)
image.plot(iyp, add=TRUE, col=rev(col.pal(length(surf.brks)-1)), zlim=zlimp)

par(oldpar)

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

#' Compute the BPS weights by convex optimization
#'
#' @param scores [matrix] \eqn{N \times K} of expected predictive density evaluations for the K models considered
#'
#' @return conv_opt [function] to perform convex optimiazion with CVXR R package
#'
CVXR_opt <- function(scores) {
    .Call(`_spBPS_CVXR_opt`, scores)
}

#' Combine subset models wiht BPS
#'
#' @param fit_list [list] K fitted model outputs composed by two elements each: first named \eqn{epd}, second named \eqn{W}
#' @param K [integer] number of folds
#' @param rp [double] percentage of observations to take into account for optimization (\code{default=1})
#'
#' @return [matrix] posterior predictive density evaluations (each columns represent a different model)
#'
BPS_combine <- function(fit_list, K, rp) {
    .Call(`_spBPS_BPS_combine`, fit_list, K, rp)
}

#' Combine subset models wiht Pseudo-BMA
#'
#' @param fit_list [list] K fitted model outputs composed by two elements each: first named \eqn{epd}, second named \eqn{W}
#'
#' @return [matrix] posterior predictive density evaluations (each columns represent a different model)
#'
BPS_PseudoBMA <- function(fit_list) {
    .Call(`_spBPS_BPS_PseudoBMA`, fit_list)
}

#' Compute the parameters for the posteriors distribution of \eqn{\beta} and \eqn{\Sigma} (i.e. updated parameters)
#'
#' @param data [list] two elements: first named \eqn{Y}, second named \eqn{X}
#' @param priors [list] priors: named \eqn{\mu_B},\eqn{V_r},\eqn{\Psi},\eqn{\nu}
#' @param coords [matrix] sample coordinates for X and Y
#' @param hyperpar [list] two elemets: first named \eqn{\alpha}, second named \eqn{\phi}
#'
#' @return [list] posterior update parameters
#'
fit_cpp_MvT <- function(data, priors, coords, hyperpar) {
    .Call(`_spBPS_fit_cpp_MvT`, data, priors, coords, hyperpar)
}

#' Sample R draws from the posterior distributions
#'
#' @param poster [list] output from \code{fit_cpp} function
#' @param R [integer] number of posterior samples
#' @param par if \code{TRUE} only \eqn{\beta} and \eqn{\Sigma} are sampled (\eqn{\omega} is omitted)
#' @param p [integer] if \code{par = TRUE}, it specifies the column number of \eqn{X}
#'
#' @return [list] posterior samples
#'
post_draws_MvT <- function(poster, R, par, p) {
    .Call(`_spBPS_post_draws_MvT`, poster, R, par, p)
}

#' Draw from the joint posterior predictive for a set of unobserved covariates
#'
#' @param data [list] two elements: first named \eqn{Y}, second named \eqn{X}
#' @param X_u [matrix] unobserved instances covariate matrix
#' @param d_u [matrix] unobserved instances distance matrix
#' @param d_us [matrix] cross-distance between unobserved and observed instances matrix
#' @param hyperpar [list] two elemets: first named \eqn{\alpha}, second named \eqn{\phi}
#' @param poster [list] output from \code{fit_cpp} function
#' @param R [integer] number of posterior predictive samples
#'
#' @return [list] posterior predictive samples
#'
r_pred_joint_MvT <- function(data, X_u, d_u, d_us, hyperpar, poster, R) {
    .Call(`_spBPS_r_pred_joint_MvT`, data, X_u, d_u, d_us, hyperpar, poster, R)
}

#' Draw from the joint posterior predictive for a set of unobserved covariates
#'
#' @param data [list] two elements: first named \eqn{Y}, second named \eqn{X}
#' @param X_u [matrix] unobserved instances covariate matrix
#' @param d_u [matrix] unobserved instances distance matrix
#' @param d_us [matrix] cross-distance between unobserved and observed instances matrix
#' @param hyperpar [list] two elemets: first named \eqn{\alpha}, second named \eqn{\phi}
#' @param poster [list] output from \code{fit_cpp} function
#' @param R [integer] number of posterior predictive samples
#'
#' @return [list] posterior predictive samples
#'
r_pred_marg_MvT <- function(data, X_u, d_u, d_us, hyperpar, poster, R) {
    .Call(`_spBPS_r_pred_marg_MvT`, data, X_u, d_u, d_us, hyperpar, poster, R)
}

#' Draw from the conditional posterior predictive for a set of unobserved covariates
#'
#' @param data [list] two elements: first named \eqn{Y}, second named \eqn{X}
#' @param X_u [matrix] unobserved instances covariate matrix
#' @param d_u [matrix] unobserved instances distance matrix
#' @param d_us [matrix] cross-distance between unobserved and observed instances matrix
#' @param hyperpar [list] two elemets: first named \eqn{\alpha}, second named \eqn{\phi}
#' @param poster [list] output from \code{fit_cpp_MvT} function
#' @param post [list] output from \code{post_draws_MvT} function
#'
#' @return [list] posterior predictive samples
#'
r_pred_cond_MvT <- function(data, X_u, d_u, d_us, hyperpar, poster, post) {
    .Call(`_spBPS_r_pred_cond_MvT`, data, X_u, d_u, d_us, hyperpar, poster, post)
}

#' Evaluate the density of a set of unobserved response with respect to the conditional posterior predictive
#'
#' @param data [list] two elements: first named \eqn{Y}, second named \eqn{X}
#' @param X_u [matrix] unobserved instances covariate matrix
#' @param Y_u [matrix] unobserved instances response matrix
#' @param d_u [matrix] unobserved instances distance matrix
#' @param d_us [matrix] cross-distance between unobserved and observed instances matrix
#' @param hyperpar [list] two elemets: first named \eqn{\alpha}, second named \eqn{\phi}
#' @param poster [list] output from \code{fit_cpp} function
#'
#' @return [double] posterior predictive density evaluation
#'
d_pred_cpp_MvT <- function(data, X_u, Y_u, d_u, d_us, hyperpar, poster) {
    .Call(`_spBPS_d_pred_cpp_MvT`, data, X_u, Y_u, d_u, d_us, hyperpar, poster)
}

#' Compute the LOOCV of the density evaluations for fixed values of the hyperparameters
#'
#' @param data [list] two elements: first named \eqn{Y}, second named \eqn{X}
#' @param priors [list] priors: named \eqn{\mu_B},\eqn{V_r},\eqn{\Psi},\eqn{\nu}
#' @param coords [matrix] sample coordinates for X and Y
#' @param hyperpar [list] two elemets: first named \eqn{\alpha}, second named \eqn{\phi}
#'
#' @return [vector] posterior predictive density evaluations
#'
dens_loocv_MvT <- function(data, priors, coords, hyperpar) {
    .Call(`_spBPS_dens_loocv_MvT`, data, priors, coords, hyperpar)
}

#' Compute the KCV of the density evaluations for fixed values of the hyperparameters
#'
#' @param data [list] two elements: first named \eqn{Y}, second named \eqn{X}
#' @param priors [list] priors: named \eqn{\mu_B},\eqn{V_r},\eqn{\Psi},\eqn{\nu}
#' @param coords [matrix] sample coordinates for X and Y
#' @param hyperpar [list] two elemets: first named \eqn{\alpha}, second named \eqn{\phi}
#' @param K [integer] number of folds
#'
#' @return [vector] posterior predictive density evaluations
#'
dens_kcv_MvT <- function(data, priors, coords, hyperpar, K) {
    .Call(`_spBPS_dens_kcv_MvT`, data, priors, coords, hyperpar, K)
}

#' Return the CV predictive density evaluations for all the model combinations
#'
#' @param data [list] two elements: first named \eqn{Y}, second named \eqn{X}
#' @param priors [list] priors: named \eqn{\mu_B},\eqn{V_r},\eqn{\Psi},\eqn{\nu}
#' @param coords [matrix] sample coordinates for X and Y
#' @param hyperpar [list] two elemets: first named \eqn{\alpha}, second named \eqn{\phi}
#' @param useKCV if \code{TRUE} K-fold cross validation is used instead of LOOCV (no \code{default})
#' @param K [integer] number of folds
#'
#' @return [matrix] posterior predictive density evaluations (each columns represent a different model)
#'
models_dens_MvT <- function(data, priors, coords, hyperpar, useKCV, K) {
    .Call(`_spBPS_models_dens_MvT`, data, priors, coords, hyperpar, useKCV, K)
}

#' Compute the BPS weights by convex optimization
#'
#' @param data [list] two elements: first named \eqn{Y}, second named \eqn{X}
#' @param priors [list] priors: named \eqn{\mu_B},\eqn{V_r},\eqn{\Psi},\eqn{\nu}
#' @param coords [matrix] sample coordinates for X and Y
#' @param hyperpar [list] two elemets: first named \eqn{\alpha}, second named \eqn{\phi}
#' @param K [integer] number of folds
#'
#' @return [matrix] posterior predictive density evaluations (each columns represent a different model)
#'
BPS_weights_MvT <- function(data, priors, coords, hyperpar, K) {
    .Call(`_spBPS_BPS_weights_MvT`, data, priors, coords, hyperpar, K)
}

#' Compute the BPS spatial prediction given a set of stacking weights
#'
#' @param data [list] two elements: first named \eqn{Y}, second named \eqn{X}
#' @param X_u [matrix] unobserved instances covariate matrix
#' @param priors [list] priors: named \eqn{\mu_B},\eqn{V_r},\eqn{\Psi},\eqn{\nu}
#' @param coords [matrix] sample coordinates for X and Y
#' @param crd_u [matrix] unboserved instances coordinates
#' @param hyperpar [list] two elemets: first named \eqn{\alpha}, second named \eqn{\phi}
#' @param W [matrix] set of stacking weights
#' @param R [integer] number of desired samples
#'
#' @return [list] BPS posterior predictive samples
#'
BPS_pred_MvT <- function(data, X_u, priors, coords, crd_u, hyperpar, W, R) {
    .Call(`_spBPS_BPS_pred_MvT`, data, X_u, priors, coords, crd_u, hyperpar, W, R)
}

#' Perform the BPS sampling from posterior and posterior predictive given a set of stacking weights
#'
#' @param data [list] two elements: first named \eqn{Y}, second named \eqn{X}
#' @param X_u [matrix] unobserved instances covariate matrix
#' @param priors [list] priors: named \eqn{\mu_B},\eqn{V_r},\eqn{\Psi},\eqn{\nu}
#' @param coords [matrix] sample coordinates for X and Y
#' @param crd_u [matrix] unboserved instances coordinates
#' @param hyperpar [list] two elemets: first named \eqn{\alpha}, second named \eqn{\phi}
#' @param W [matrix] set of stacking weights
#' @param R [integer] number of desired samples
#'
#' @return [list] BPS posterior predictive samples
#'
BPS_post_MvT <- function(data, X_u, priors, coords, crd_u, hyperpar, W, R) {
    .Call(`_spBPS_BPS_post_MvT`, data, X_u, priors, coords, crd_u, hyperpar, W, R)
}

#' Compute the BPS posterior samples given a set of stacking weights
#'
#' @param data [list] two elements: first named \eqn{Y}, second named \eqn{X}
#' @param priors [list] priors: named \eqn{\mu_B},\eqn{V_r},\eqn{\Psi},\eqn{\nu}
#' @param coords [matrix] sample coordinates for X and Y
#' @param hyperpar [list] two elemets: first named \eqn{\alpha}, second named \eqn{\phi}
#' @param W [matrix] set of stacking weights
#' @param R [integer] number of desired samples
#' @param par if \code{TRUE} only \eqn{\beta} and \eqn{\Sigma} are sampled (\eqn{\omega} is omitted)
#'
#' @return [matrix] BPS posterior samples
#'
BPS_postdraws_MvT <- function(data, priors, coords, hyperpar, W, R, par) {
    .Call(`_spBPS_BPS_postdraws_MvT`, data, priors, coords, hyperpar, W, R, par)
}

#' @keywords internal
BPS_post_all_subsets_v2 <- function(Y_list, X_list, crd_list, W_local_list, subset_ind, priors, hyperpar, n_cores) {
    .Call(`_spBPS_BPS_post_all_subsets_v2`, Y_list, X_list, crd_list, W_local_list, subset_ind, priors, hyperpar, n_cores)
}

#' @keywords internal
BPS_pred_streaming_v2 <- function(Y_list, X_list, crd_list, W_local_list, subset_ind, crd_u, X_u, priors, hyperpar, pred_batch_size, n_cores, return_summary, prob_lo, prob_hi) {
    .Call(`_spBPS_BPS_pred_streaming_v2`, Y_list, X_list, crd_list, W_local_list, subset_ind, crd_u, X_u, priors, hyperpar, pred_batch_size, n_cores, return_summary, prob_lo, prob_hi)
}

#' @keywords internal
fit_cpp_MvT_v2 <- function(data, priors, coords, hyperpar, iVr_pre) {
    .Call(`_spBPS_fit_cpp_MvT_v2`, data, priors, coords, hyperpar, iVr_pre)
}

#' @keywords internal
post_draws_MvT_v2 <- function(poster, R, par, p) {
    .Call(`_spBPS_post_draws_MvT_v2`, poster, R, par, p)
}

#' @keywords internal
d_pred_cpp_MvT_v2 <- function(data, X_u, Y_u, cross_d, hyperpar, poster, Lc_Psi) {
    .Call(`_spBPS_d_pred_cpp_MvT_v2`, data, X_u, Y_u, cross_d, hyperpar, poster, Lc_Psi)
}

#' @keywords internal
dens_kcv_MvT_v2 <- function(data, priors, coords, hyperpar, K) {
    .Call(`_spBPS_dens_kcv_MvT_v2`, data, priors, coords, hyperpar, K)
}

#' @keywords internal
models_dens_MvT_v2 <- function(data, priors, coords, hyperpar, K) {
    .Call(`_spBPS_models_dens_MvT_v2`, data, priors, coords, hyperpar, K)
}

#' @keywords internal
BPS_weights_MvT_v2 <- function(data, priors, coords, hyperpar, K) {
    .Call(`_spBPS_BPS_weights_MvT_v2`, data, priors, coords, hyperpar, K)
}

#' @keywords internal
BPS_combine_v2 <- function(fit_list, K, rp) {
    .Call(`_spBPS_BPS_combine_v2`, fit_list, K, rp)
}

#' @keywords internal
compute_all_local_weights_v2 <- function(Y_list, X_list, crd_list, priors, hyperpar, cv_folds, n_cores) {
    .Call(`_spBPS_compute_all_local_weights_v2`, Y_list, X_list, crd_list, priors, hyperpar, cv_folds, n_cores)
}

#' @keywords internal
BPS_post_MvT_v2 <- function(data, X_u, priors, coords, crd_u, hyperpar, W, R, d_u, Rphi_u_list, n_cores) {
    .Call(`_spBPS_BPS_post_MvT_v2`, data, X_u, priors, coords, crd_u, hyperpar, W, R, d_u, Rphi_u_list, n_cores)
}

#' @keywords internal
BPS_post_MvT_v2_batch <- function(data, X_u, priors, coords, crd_u, hyperpar, W, R, d_u, Rphi_u_list, n_cores, pred_batch_size) {
    .Call(`_spBPS_BPS_post_MvT_v2_batch`, data, X_u, priors, coords, crd_u, hyperpar, W, R, d_u, Rphi_u_list, n_cores, pred_batch_size)
}

#' @keywords internal
BPS_pred_all_subsets_v2 <- function(Y_list, X_list, crd_list, W_local_list, subset_ind, d_u, Rphi_u_list, X_u, crd_u, priors, hyperpar, pred_batch_size, n_cores) {
    .Call(`_spBPS_BPS_pred_all_subsets_v2`, Y_list, X_list, crd_list, W_local_list, subset_ind, d_u, Rphi_u_list, X_u, crd_u, priors, hyperpar, pred_batch_size, n_cores)
}

#' @keywords internal
BPS_postdraws_MvT_v2 <- function(data, priors, coords, hyperpar, W, R, par) {
    .Call(`_spBPS_BPS_postdraws_MvT_v2`, data, priors, coords, hyperpar, W, R, par)
}

#' @keywords internal
BPS_PseudoBMA_v2 <- function(fit_list, n_cores = 1L) {
    .Call(`_spBPS_BPS_PseudoBMA_v2`, fit_list, n_cores)
}

#' @keywords internal
dmatrix_t_test <- function(X, Lambda, SigmaR, SigmaC, nu, log_d = FALSE) {
    .Call(`_spBPS_dmatrix_t_test`, X, Lambda, SigmaR, SigmaC, nu, log_d)
}

#' Compute the Euclidean distance matrix
#'
#' @param X [matrix] (tipically of \eqn{N} coordindates on \eqn{\mathbb{R}^2} )
#'
#' @return [matrix] distance matrix of the elements of \eqn{X}
#'
#'
arma_dist <- function(X) {
    .Call(`_spBPS_arma_dist`, X)
}

#' Build a grid from two vector (i.e. equivalent to \code{expand.grid()} in \code{R})
#'
#' @param x [vector] first vector of numeric elements
#' @param y [vector] second vector of numeric elements
#'
#' @return [matrix] expanded grid of combinations
#'
#'
expand_grid_cpp <- function(x, y) {
    .Call(`_spBPS_expand_grid_cpp`, x, y)
}

#' Function to sample integers (index)
#'
#' @param size [integer] dimension of the set to sample
#' @param length [integer] number of elements to sample
#' @param p [vector] sampling probabilities
#'
#' @return [vector] sample of integers
#'
sample_index <- function(size, length, p) {
    .Call(`_spBPS_sample_index`, size, length, p)
}

#' Function to subset data for meta-analysis
#'
#' @param data [list] three elements: first named \eqn{Y}, second named \eqn{X}, third named \eqn{crd}
#' @param K [integer] number of desired subsets
#'
#' @return [list] subsets of data, and the set of indexes
#'
#' @export
subset_data <- function(data, K) {
    .Call(`_spBPS_subset_data`, data, K)
}

#' Function to subset data for meta-analysis
#'
#' @param mat [matrix] not-symmetric matrix
#'
#' @return [matrix] symmetric matrix (lower triangular of \code{mat} is used)
#'
forceSymmetry_cpp <- function(mat) {
    .Call(`_spBPS_forceSymmetry_cpp`, mat)
}

utils::globalVariables("i")

#' Unified spatial BPS workflow (multivariate path, works for q = 1)
#'
#' @description
#' Orchestrates subsetting, local stacking weight estimation, global stacking
#' combination, and optional posterior or predictive simulation using the
#' multivariate Student-t spatial model. Works for both multivariate outcomes
#' and the univariate case via `q = 1`.
#'
#' @param data List with matrices `Y` (response) and `X` (covariates).
#' @param priors List of priors for the multivariate model (`mu_B`, `V_r`,
#'   `Psi`, `nu`).
#' @param coords Matrix of observation coordinates.
#' @param hyperpar List with elements `alpha` and `phi` (vectors allowed).
#' @param subset_size Target subset size when `K` is not provided. Default 500.
#' @param K Optional number of subsets. When `NULL`, computed as
#'   `ceiling(nrow(Y) / subset_size)` and lower-bounded at 1.
#' @param cv_folds Number of folds for local cross-validation (default 5).
#' @param rp Fraction of rows used when recomputing global stacking weights
#'   (passed to `BPS_combine`). Ignored when `combine_method = "pseudoBMA"`.
#' @param combine_method Choose between Bayesian Predictive Stacking (`"bps"`)
#'   or pseudo-BMA (`"pseudoBMA"`) for combining subsets.
#' @param draws Number of joint posterior/predictive draws to return (0 to
#'   skip). When positive, `newdata` must be supplied because draws are obtained
#'   via `BPS_post_MvT` which jointly samples posterior and predictive.
#' @param newdata Optional list with `X` and `coords` for prediction locations;
#'   required when either draw count is positive.
#' @param include_latent Logical; if `TRUE`, posterior draws include latent
#'   processes.
#' @param cores Optional integer; when >1 a parallel backend is registered
#'   internally via `doParallel::registerDoParallel(cores)` for the fit and draw
#'   loops. When `NULL`, the existing foreach backend (if any) is used.
#'
#' @return List with components `subsets`, `weights_global`, `weights_local`,
#'   `epd`, and optional `posterior` and `predictive` draws.
#'
#' @examples
#' \donttest{
#' n <- 1000
#' p <- 2
#' q <- 1
#'
#' Y <- matrix(rnorm(n*q), ncol = q)
#' X <- matrix(rnorm(n*p), ncol = p)
#' coords <- matrix(runif(n*2), ncol = 2)
#'
#' data <- list(Y = Y, X = X)
#' priors <- list(mu_B = matrix(0, nrow = p, ncol = q),
#'                              V_r = diag(10, p),
#'                              Psi = diag(1, q),
#'                              nu = 3)
#' hyperpar <- list(alpha = 0.5, phi = 1)
#' subset_size <- 200
#'
#' res <- spBPS(data, priors, coords, hyperpar, subset_size = subset_size)
#'
#' }
#'
spBPS_old <- function(data,
                      priors,
                      coords,
                      hyperpar,
                      subset_size = 500L,
                      K = NULL,
                      cv_folds = 5L,
                      rp = 1,
                      combine_method = c("bps", "pseudoBMA"),
                      draws = 0L,
                      newdata = NULL,
                      include_latent = FALSE,
                      cores = NULL) {

  cat("\n====================================================\n")
  cat("         Welcome to spBPS Bayesian Engine\n")
  cat("====================================================\n\n")

  combine_method <- match.arg(combine_method)

  map_parallel <- function(indices, fn, exports = NULL, packages = NULL) {
    if (requireNamespace("foreach", quietly = TRUE)) {
      `%dopar%` <- foreach::`%dopar%`
      workers <- tryCatch(foreach::getDoParWorkers(), error = function(e) 1L)
      if (!is.null(workers) && workers > 1L) {
        export_vars <- unique(c(exports, ls(envir = parent.frame()), ls(envir = environment(fn))))
        return(foreach::foreach(i = indices,
                                 .export = export_vars,
                                 .packages = packages) %dopar% fn(i))
      }
    }
    lapply(indices, fn)
  }

  # optional internal backend
  if (!is.null(cores) && cores > 1L) {
    cat("Preparing parallel backend with", cores, "cores... \n\n")
    if (!requireNamespace("doParallel", quietly = TRUE)) {
      stop("Package 'doParallel' is required when 'cores' > 1.")
    }
    cl <- parallel::makeCluster(cores)
    doParallel::registerDoParallel(cl)
    on.exit({
      try(parallel::stopCluster(cl), silent = TRUE)
      foreach::registerDoSEQ()
    }, add = TRUE)
  }

  Y <- data$Y
  if (is.null(dim(Y))) {
    Y <- matrix(Y, ncol = 1L)
  }
  X <- data$X
  n <- nrow(Y)

  if (is.null(K)) {
    K <- max(1L, min(n, as.integer(ceiling(n / subset_size))))
  }

  # Subset data once using the provided coordinates
  cat("Pritioning data into K =", K, "subsets ... \n\n")
  subsets <- spBPS::subset_data(data = list(Y = Y, X = X, crd = coords), K = K)
  # Local models: compute weights and epd per subset
  J <- nrow(expand_grid_cpp(hyperpar$alpha, hyperpar$phi))
  cat("Computing local stacking weights over J =", J,"models ...\n")
  fit_list <- map_parallel(seq_len(K), function(i) {
    Yi <- subsets$Y_list[[i]]
    Xi <- subsets$X_list[[i]]
    crd_i <- subsets$crd_list[[i]]
    p <- ncol(Xi)
    q <- ncol(Yi)

    local <- suppressWarnings(BPS_weights_MvT(
      data = list(Y = Yi, X = Xi),
      priors = priors,
      coords = crd_i,
      hyperpar = hyperpar,
      K = cv_folds
    ))

    list(local$epd, local$W)
  }, exports = c("subsets", "priors", "hyperpar", "cv_folds"), packages = "spBPS")
  cat("Local weights computed.\n\n")

  # Global combination
  cat("Computing global stacking weights over K =", K,"partitions ...\n")
  if (combine_method == "bps") {
    comb <- suppressWarnings(BPS_combine(fit_list = fit_list, K = K, rp = rp))
  } else {
    comb <- BPS_PseudoBMA(fit_list = fit_list)
  }

  W_global <- comb$W
  W_local <- comb$W_list
  cat("Global weights computed.\n\n")

  out <- list(
    subsets = subsets,
    weights_global = W_global,
    weights_local = W_local,
    epd = lapply(fit_list, `[[`, 1)
  )

  # Joint posterior + predictive draws via BPS_post_MvT (requires newdata when draws requested)
  draw_total <- draws
  if (draw_total > 0) {

    if (is.null(newdata)) {

      cat("Posterior Sampling (R =",draws, "draws) ...\n")
      subset_ind <- sample.int(K, draw_total, replace = TRUE, prob = W_global)
      draw_list <- map_parallel(seq_len(draw_total), function(r) {
        idx <- subset_ind[r]
        Yi <- subsets$Y_list[[idx]]
        Xi <- subsets$X_list[[idx]]
        crd_i <- subsets$crd_list[[idx]]
        res <- BPS_postdraws_MvT(
          data = list(Y = Yi, X = Xi),
          priors = priors,
          coords = crd_i,
          hyperpar = hyperpar,
          W = W_local[[idx]],
          R = 1,
          par = !isTRUE(include_latent)
        )
        list(post = res[[1]][[1]])
      }, exports = c("subsets", "priors", "hyperpar", "W_local", "newdata"), packages = "spBPS")

      out$posterior <- lapply(draw_list, `[[`, "post")
      cat("Posterior sampling completed.\n\n")

    } else {

    cat("Posterior & Predictive Sampling (R =",draws, "draws) ...\n")
    subset_ind <- sample.int(K, draw_total, replace = TRUE, prob = W_global)
    draw_list <- map_parallel(seq_len(draw_total), function(r) {
      idx <- subset_ind[r]
      Yi <- subsets$Y_list[[idx]]
      Xi <- subsets$X_list[[idx]]
      crd_i <- subsets$crd_list[[idx]]
      res <- BPS_post_MvT(
        data = list(Y = Yi, X = Xi),
        X_u = newdata$X,
        priors = priors,
        coords = crd_i,
        crd_u = newdata$coords,
        hyperpar = hyperpar,
        W = W_local[[idx]],
        R = 1
      )
      list(post = res$Post[[1]], pred = res$Pred[[1]])
    }, exports = c("subsets", "priors", "hyperpar", "W_local", "newdata"), packages = "spBPS")

    out$posterior <- lapply(draw_list, `[[`, "post")
    out$predictive <- lapply(draw_list, `[[`, "pred")
    cat("Posterior & Predictive sampling completed.\n\n")
  }

  }

  cat("====================================================\n")
  cat("     spBPS pipeline completed successfully!\n")
  cat("====================================================\n\n")

  return(out)
}

---
title: "Double Bayesian Predictive Stacking for Spatial Analysis - Tutotial"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Double Bayesian Predictive Stacking for Spatial Analysis - Tutotial}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
bibliography: references.bib
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

We provide a brief tutorial of the `spBPS` package. Here we shows the implementation of the Double Bayesian Predictive Stacking on synthetically univariate generated data. In particular, we focus on parallel computing using the packages `parallel`, `doParallel`; but it is not mandatory: it suffices to make it sequential. For any further details please refer to [@presicce2024]. More examples, for multivariate data, are available in documentation, and functions help.

```{r setup}
library(spBPS)
```

### Working packages
```{r, warning=FALSE, message=F}
library(foreach)
library(parallel)
library(doParallel)
library(tictoc)
library(MBA)
library(classInt)
library(RColorBrewer)
library(sp)
library(fields)
library(mvnfast)
library(abind)
```


### Data generation
We generate data from the model detailed in Equation (2.4) [@presicce2024], over a unit square.
```{r, results=F}
# dimensions
n <- 1000
u <- 250
p <- 2
q <- 1

# parameters
B <- c(-0.75, 1.85)
tau2 <- 0.25
sigma2 <- 1
delta <- tau2/sigma2
phi <- 4

set.seed(4-8-15-16-23-42)
# generate sintethic data
crd <- matrix(runif((n+u) * 2), ncol = 2)
X_or <- cbind(rep(1, n+u), matrix(runif((p-1)*(n+u)), ncol = (p-1)))
D <- spBPS:::arma_dist(crd)
Rphi <- exp(-phi * D)
W_or <- matrix(0, n+u) + mniw::rmNorm(1, rep(0, n+u), sigma2*Rphi)
Y_or <- X_or %*% B + W_or + mniw::rmNorm(1, rep(0, n+u), diag(delta*sigma2, n+u))

# train data
crd_s <- crd[1:n, ]
X <- X_or[1:n, ]
W <- W_or[1:n, ]
Y <- Y_or[1:n, ]

# prediction data
crd_u <- crd[-(1:n), ]
X_u <- X_or[-(1:n), ]
W_u <- W_or[-(1:n), ]
Y_u <- Y_or[-(1:n), ]
```

### Setting priors and hyperparameters
```{r, results=F}
# priors 
priors <- list(mu_B = matrix(0, nrow = p, ncol = q),
               V_r = diag(10, p),
               Psi = diag(1, q),
               nu = 3)

# hyperparameters values
alfa_seq <- c(0.7, 0.8, 0.9)
phi_seq <- c(3, 4, 5)
hyperpar <- list(alpha = alfa_seq, phi = phi_seq)
```

### Setting dimensions
```{r, results=F}
# subset dimension
subset_size <- 100

# number of posterior draws
R <- 200

# number of computational cores
n_core <- 1
```

### Double BPS parallel fit
Parallel implementation, exploiting 1 computing core.
```{r, results=F}
out <- spBPS(data = list(Y = Y, X = X),
      priors = priors,
      coords = crd_s,
      hyperpar = hyperpar,
      subset_size = subset_size,
      combine_method = "bps",
      draws = R,
      newdata = list(X = X_u, coords = crd_u),
      n_cores = n_core)
```

### Results collection
```{r}
# statistics computations W
pred_mat_W <- out$predictive$Wu
post_mean_W <- apply(pred_mat_W, c(1,2), mean)
post_qnt_W <- apply(pred_mat_W, c(1,2), quantile, c(0.025, 0.975))

# Empirical coverage for W
coverage_W <- mean(W_u >= post_qnt_W[1,,1] & W_u <= post_qnt_W[2,,1])
cat("Empirical coverage for Spatial process:", round(coverage_W, 3),"\n")

# statistics computations Y
pred_mat_Y <- out$predictive$Yu
post_mean_Y <- apply(pred_mat_Y, c(1,2), mean)
post_qnt_Y <- apply(pred_mat_Y, c(1,2), quantile, c(0.025, 0.975))

# Empirical coverage for Y
coverage_Y <- mean(Y_u >= post_qnt_Y[1,,1] & Y_u <= post_qnt_Y[2,,1])
cat("Empirical coverage for Response:", round(coverage_Y, 3),"\n")

# Root Mean Square Prediction Error
rmspe_W <- sqrt( mean( (W_u - post_mean_W)^2 ) )
rmspe_Y <- sqrt( mean( (Y_u - post_mean_Y)^2 ) )
cat("RMSPE for Spatial process:", round(rmspe_W, 3), "\n")
cat("RMSPE for Response:", round(rmspe_Y, 3), "\n")
```

### Plot results
```{r, results=F, echo=F}
# True spatial process surface interpolation
h <- 12
surf.W <- MBA::mba.surf(cbind(crd_s, W), no.X = 500, no.Y = 500,
                        exten = TRUE, sp = TRUE, h = h)$xyz.est
surf.brks <- classIntervals(surf.W$z, 100, 'pretty')$brks
col.pal <- colorRampPalette(brewer.pal(11,'RdBu')[11:1])
xlim <- c(0, 1)
zlim <- range(surf.W$z)

# image for plot
iw <- as.image.SpatialGridDataFrame(surf.W)

# BPS surfaces interpolation
h <- 12
surf.Wp <- MBA::mba.surf(cbind(crd_u, post_mean_W), no.X = 500, no.Y = 500,
                         exten = TRUE, sp = TRUE, h = h)$xyz.est
zlimp <- range(surf.Wp$z)

# image for plot
iwp <- as.image.SpatialGridDataFrame(surf.Wp)
```

```{r, echo=F, fig.dim = c(7.25, 4)}
# Plotting
oldpar <- par(no.readonly = TRUE)
par(mfrow = c(1, 2))

plot(crd, type="n", cex=0.5, xlim=xlim, axes=FALSE, ylab="Northing", xlab="Easting",
     main="Spatial process")
axis(2, las=1)
axis(1)
image.plot(iw, add=TRUE, col=rev(col.pal(length(surf.brks)-1)), zlim=zlim)

plot(crd, type="n", cex=0.5, xlim=xlim, axes=F, ylab="Northing", xlab="Easting")
title(main="Spatial process (Prediction)")
mtext(side = 3, paste("RMSPE :", round(rmspe_W, 3)))
axis(2, las=1)
axis(1)
image.plot(iwp, add=TRUE, col=rev(col.pal(length(surf.brks)-1)), zlim=zlimp)

par(oldpar)
```

```{r, results=F, echo=F}
# True response surface interpolation
h <- 12
surf.Y <- MBA::mba.surf(cbind(crd_s, Y), no.X = 500, no.Y = 500,
                        exten = TRUE, sp = TRUE, h = h)$xyz.est
surf.brks <- classIntervals(surf.Y$z, 100, 'pretty')$brks
col.pal <- colorRampPalette(brewer.pal(11,'RdBu')[11:1])
xlim <- c(0, 1)
zlim <- range(surf.Y$z)

# image for plot
iy <- as.image.SpatialGridDataFrame(surf.Y)

# BPS surfaces interpolation
h <- 12
surf.Yp <- MBA::mba.surf(cbind(crd_u, post_mean_Y), no.X = 500, no.Y = 500,
                         exten = TRUE, sp = TRUE, h = h)$xyz.est
zlimp <- range(surf.Yp$z)

# image for plot
iyp <- as.image.SpatialGridDataFrame(surf.Yp)
```

```{r, echo=F, fig.dim = c(7.25, 4)}
# Plotting
oldpar <- par(no.readonly = TRUE)
par(mfrow = c(1, 2))

plot(crd, type="n", cex=0.5, xlim=xlim, axes=FALSE, ylab="Northing", xlab="Easting",
     main="Response")
axis(2, las=1)
axis(1)
image.plot(iy, add=TRUE, col=rev(col.pal(length(surf.brks)-1)), zlim=zlim)

plot(crd, type="n", cex=0.5, xlim=xlim, axes=F, ylab="Northing", xlab="Easting")
title(main="Response (Prediction)")
mtext(side = 3, paste("RMSPE :", round(rmspe_Y, 3)))
axis(2, las=1)
axis(1)
image.plot(iyp, add=TRUE, col=rev(col.pal(length(surf.brks)-1)), zlim=zlimp)

par(oldpar)
```


------

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{d_pred_cpp_MvT}
\alias{d_pred_cpp_MvT}
\title{Evaluate the density of a set of unobserved response with respect to the conditional posterior predictive}
\usage{
d_pred_cpp_MvT(data, X_u, Y_u, d_u, d_us, hyperpar, poster)
}
\arguments{
\item{data}{\link{list} two elements: first named \eqn{Y}, second named \eqn{X}}

\item{X_u}{\link{matrix} unobserved instances covariate matrix}

\item{Y_u}{\link{matrix} unobserved instances response matrix}

\item{d_u}{\link{matrix} unobserved instances distance matrix}

\item{d_us}{\link{matrix} cross-distance between unobserved and observed instances matrix}

\item{hyperpar}{\link{list} two elemets: first named \eqn{\alpha}, second named \eqn{\phi}}

\item{poster}{\link{list} output from \code{fit_cpp} function}
}
\value{
\link{double} posterior predictive density evaluation
}
\description{
Evaluate the density of a set of unobserved response with respect to the conditional posterior predictive
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/spBPS.R
\name{spBPS_old}
\alias{spBPS_old}
\title{Unified spatial BPS workflow (multivariate path, works for q = 1)}
\usage{
spBPS_old(
  data,
  priors,
  coords,
  hyperpar,
  subset_size = 500L,
  K = NULL,
  cv_folds = 5L,
  rp = 1,
  combine_method = c("bps", "pseudoBMA"),
  draws = 0L,
  newdata = NULL,
  include_latent = FALSE,
  cores = NULL
)
}
\arguments{
\item{data}{List with matrices \code{Y} (response) and \code{X} (covariates).}

\item{priors}{List of priors for the multivariate model (\code{mu_B}, \code{V_r},
\code{Psi}, \code{nu}).}

\item{coords}{Matrix of observation coordinates.}

\item{hyperpar}{List with elements \code{alpha} and \code{phi} (vectors allowed).}

\item{subset_size}{Target subset size when \code{K} is not provided. Default 500.}

\item{K}{Optional number of subsets. When \code{NULL}, computed as
\code{ceiling(nrow(Y) / subset_size)} and lower-bounded at 1.}

\item{cv_folds}{Number of folds for local cross-validation (default 5).}

\item{rp}{Fraction of rows used when recomputing global stacking weights
(passed to \code{BPS_combine}). Ignored when \code{combine_method = "pseudoBMA"}.}

\item{combine_method}{Choose between Bayesian Predictive Stacking (\code{"bps"})
or pseudo-BMA (\code{"pseudoBMA"}) for combining subsets.}

\item{draws}{Number of joint posterior/predictive draws to return (0 to
skip). When positive, \code{newdata} must be supplied because draws are obtained
via \code{BPS_post_MvT} which jointly samples posterior and predictive.}

\item{newdata}{Optional list with \code{X} and \code{coords} for prediction locations;
required when either draw count is positive.}

\item{include_latent}{Logical; if \code{TRUE}, posterior draws include latent
processes.}

\item{cores}{Optional integer; when >1 a parallel backend is registered
internally via \code{doParallel::registerDoParallel(cores)} for the fit and draw
loops. When \code{NULL}, the existing foreach backend (if any) is used.}
}
\value{
List with components \code{subsets}, \code{weights_global}, \code{weights_local},
\code{epd}, and optional \code{posterior} and \code{predictive} draws.
}
\description{
Orchestrates subsetting, local stacking weight estimation, global stacking
combination, and optional posterior or predictive simulation using the
multivariate Student-t spatial model. Works for both multivariate outcomes
and the univariate case via \code{q = 1}.
}
\examples{
\donttest{
n <- 1000
p <- 2
q <- 1

Y <- matrix(rnorm(n*q), ncol = q)
X <- matrix(rnorm(n*p), ncol = p)
coords <- matrix(runif(n*2), ncol = 2)

data <- list(Y = Y, X = X)
priors <- list(mu_B = matrix(0, nrow = p, ncol = q),
                             V_r = diag(10, p),
                             Psi = diag(1, q),
                             nu = 3)
hyperpar <- list(alpha = 0.5, phi = 1)
subset_size <- 200

res <- spBPS(data, priors, coords, hyperpar, subset_size = subset_size)

}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{expand_grid_cpp}
\alias{expand_grid_cpp}
\title{Build a grid from two vector (i.e. equivalent to \code{expand.grid()} in \code{R})}
\usage{
expand_grid_cpp(x, y)
}
\arguments{
\item{x}{\link{vector} first vector of numeric elements}

\item{y}{\link{vector} second vector of numeric elements}
}
\value{
\link{matrix} expanded grid of combinations
}
\description{
Build a grid from two vector (i.e. equivalent to \code{expand.grid()} in \code{R})
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{forceSymmetry_cpp}
\alias{forceSymmetry_cpp}
\title{Function to subset data for meta-analysis}
\usage{
forceSymmetry_cpp(mat)
}
\arguments{
\item{mat}{\link{matrix} not-symmetric matrix}
}
\value{
\link{matrix} symmetric matrix (lower triangular of \code{mat} is used)
}
\description{
Function to subset data for meta-analysis
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{CVXR_opt}
\alias{CVXR_opt}
\title{Compute the BPS weights by convex optimization}
\usage{
CVXR_opt(scores)
}
\arguments{
\item{scores}{\link{matrix} \eqn{N \times K} of expected predictive density evaluations for the K models considered}
}
\value{
conv_opt \link{function} to perform convex optimiazion with CVXR R package
}
\description{
Compute the BPS weights by convex optimization
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{r_pred_joint_MvT}
\alias{r_pred_joint_MvT}
\title{Draw from the joint posterior predictive for a set of unobserved covariates}
\usage{
r_pred_joint_MvT(data, X_u, d_u, d_us, hyperpar, poster, R)
}
\arguments{
\item{data}{\link{list} two elements: first named \eqn{Y}, second named \eqn{X}}

\item{X_u}{\link{matrix} unobserved instances covariate matrix}

\item{d_u}{\link{matrix} unobserved instances distance matrix}

\item{d_us}{\link{matrix} cross-distance between unobserved and observed instances matrix}

\item{hyperpar}{\link{list} two elemets: first named \eqn{\alpha}, second named \eqn{\phi}}

\item{poster}{\link{list} output from \code{fit_cpp} function}

\item{R}{\link{integer} number of posterior predictive samples}
}
\value{
\link{list} posterior predictive samples
}
\description{
Draw from the joint posterior predictive for a set of unobserved covariates
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{sample_index}
\alias{sample_index}
\title{Function to sample integers (index)}
\usage{
sample_index(size, length, p)
}
\arguments{
\item{size}{\link{integer} dimension of the set to sample}

\item{length}{\link{integer} number of elements to sample}

\item{p}{\link{vector} sampling probabilities}
}
\value{
\link{vector} sample of integers
}
\description{
Function to sample integers (index)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{subset_data}
\alias{subset_data}
\title{Function to subset data for meta-analysis}
\usage{
subset_data(data, K)
}
\arguments{
\item{data}{\link{list} three elements: first named \eqn{Y}, second named \eqn{X}, third named \eqn{crd}}

\item{K}{\link{integer} number of desired subsets}
}
\value{
\link{list} subsets of data, and the set of indexes
}
\description{
Function to subset data for meta-analysis
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/spBPS_new.R
\name{spBPS}
\alias{spBPS}
\title{Unified spatial BPS workflow (multivariate path, works for q = 1)}
\usage{
spBPS(
  data,
  priors,
  coords,
  hyperpar,
  subset_size = 500L,
  K = NULL,
  cv_folds = 5L,
  rp = 1,
  combine_method = c("bps", "pseudoBMA"),
  draws = 0L,
  newdata = NULL,
  include_latent = FALSE,
  n_cores = 1L,
  pred_batch_size = 200L
)
}
\arguments{
\item{data}{List with matrices \code{Y} (response, n x q) and \code{X} (covariates, n x p).}

\item{priors}{List of priors for the multivariate model: \code{mu_B} (p x q mean
matrix), \code{V_r} (p x p covariance), \code{Psi} (q x q scale), \code{nu} (degrees of
freedom).}

\item{coords}{Matrix of observation coordinates (n x d).}

\item{hyperpar}{List with elements \code{alpha} and \code{phi} (vectors allowed); defines
the grid of models over which stacking weights are computed.}

\item{subset_size}{Target subset size when \code{K} is not provided. Default 500.}

\item{K}{Optional number of subsets. When \code{NULL}, computed as
\code{ceiling(nrow(Y) / subset_size)} and lower-bounded at 1.}

\item{cv_folds}{Number of folds for local cross-validation. Default 5.}

\item{rp}{Fraction of rows used when recomputing global stacking weights
(passed to \code{BPS_combine}). Ignored when \code{combine_method = "pseudoBMA"}.}

\item{combine_method}{Global combination method: Bayesian Predictive Stacking
(\code{"bps"}) or pseudo-BMA (\code{"pseudoBMA"}).}

\item{draws}{Number of posterior/predictive draws to return (0 to skip
sampling). When positive and \code{newdata} is supplied, joint posterior and
predictive draws are returned. When positive and \code{newdata} is \code{NULL},
only posterior draws (beta, sigma) are returned.}

\item{newdata}{Optional list with \code{X} (u x p) and \code{coords} (u x d) for
prediction locations. Required when \code{draws > 0} and predictions are desired.
For large \code{u}, use \code{pred_batch_size} to control memory usage.}

\item{include_latent}{Unused; kept for compatibility.}

\item{n_cores}{Number of cores for parallel computation. Controls both the
local weight estimation (parallel over K subsets) and predictive sampling.
Default 1.}

\item{pred_batch_size}{Batch size for streaming prediction. Controls the
maximum number of prediction sites processed at once, hence the peak memory
usage. Default 200. Rule of thumb: peak RAM (MB) ? batch_size x q x draws x 8 / 1e6.
Set \code{NULL} for automatic selection (min(200, u)).}
}
\value{
Object of class \code{"spBPS"} ? a list with components:
\describe{
\item{subsets}{Partition information: Y_list, X_list, crd_list.}
\item{weights_global}{K-vector of global stacking weights.}
\item{weights_local}{K-list of local stacking weight vectors.}
\item{epd}{K-list of n_k x J log-density matrices.}
\item{priors, hyperpar}{Stored for use by \code{\link{predict.spBPS}}.}
\item{timings}{Named numeric vector: fitting, combination, sampling (seconds).}
\item{posterior}{(if draws > 0) List with three-dimensional arrays:
\code{beta} (p x q x R), \code{sigma} (q x q x R), \code{model} (R).}
\item{predictive}{(if draws > 0 and newdata supplied) List with
three-dimensional arrays: \code{Wu}, \code{Yu}, \code{MY} (each u x q x R).}
}
}
\description{
Orchestrates subsetting, local stacking weight estimation, global stacking combination, and optional posterior or predictive simulation using Double Bayesian Predictive Stacking for latent spatial regression. Works for both multivariate outcomes and the univariate case via \code{q = 1}.
}
\examples{
\donttest{
n <- 1000
p <- 2
q <- 1

Y <- matrix(rnorm(n * q), ncol = q)
X <- matrix(rnorm(n * p), ncol = p)
coords <- matrix(runif(n * 2), ncol = 2)

data    <- list(Y = Y, X = X)
priors  <- list(mu_B = matrix(0, nrow = p, ncol = q),
                V_r  = diag(10, p),
                Psi  = diag(1, q),
                nu   = 3)
hyperpar <- list(alpha = 0.5, phi = 1)

res <- spBPS(data, priors, coords, hyperpar, subset_size = 200)
}

}
\seealso{
\code{\link{predict.spBPS}} for generating predictions on new data
after fitting.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{arma_dist}
\alias{arma_dist}
\title{Compute the Euclidean distance matrix}
\usage{
arma_dist(X)
}
\arguments{
\item{X}{\link{matrix} (tipically of \eqn{N} coordindates on \eqn{\mathbb{R}^2} )}
}
\value{
\link{matrix} distance matrix of the elements of \eqn{X}
}
\description{
Compute the Euclidean distance matrix
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/spBPS-package.R
\docType{package}
\name{spBPS-package}
\alias{spBPS-package}
\title{spBPS: Bayesian Predictive Stacking for Scalable Geospatial Transfer Learning}
\description{
\if{html}{\figure{logo.png}{options: style='float: right' alt='logo' width='120'}}

Provides functions for Bayesian Predictive Stacking within the Bayesian transfer learning framework for geospatial artificial systems, as introduced in "Bayesian Transfer Learning for Artificially Intelligent Geospatial Systems: A Predictive Stacking Approach" (Presicce and Banerjee, 2025) \doi{10.48550/arXiv.2410.09504}. This methodology enables efficient Bayesian geostatistical modeling, utilizing predictive stacking to improve inference across spatial datasets. The core functions leverage 'C++' for high-performance computation, making the framework well-suited for large-scale spatial data analysis in parallel and distributed computing environments. Designed for scalability, it allows seamless application in computationally demanding scenarios.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://lucapresicce.github.io/spBPS/}
}

}
\author{
\strong{Maintainer}: Luca Presicce \email{l.presicce@campus.unimib.it} (\href{https://orcid.org/0009-0005-7062-3523}{ORCID})

Authors:
\itemize{
  \item Sudipto Banerjee
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{BPS_postdraws_MvT}
\alias{BPS_postdraws_MvT}
\title{Compute the BPS posterior samples given a set of stacking weights}
\usage{
BPS_postdraws_MvT(data, priors, coords, hyperpar, W, R, par)
}
\arguments{
\item{data}{\link{list} two elements: first named \eqn{Y}, second named \eqn{X}}

\item{priors}{\link{list} priors: named \eqn{\mu_B},\eqn{V_r},\eqn{\Psi},\eqn{\nu}}

\item{coords}{\link{matrix} sample coordinates for X and Y}

\item{hyperpar}{\link{list} two elemets: first named \eqn{\alpha}, second named \eqn{\phi}}

\item{W}{\link{matrix} set of stacking weights}

\item{R}{\link{integer} number of desired samples}

\item{par}{if \code{TRUE} only \eqn{\beta} and \eqn{\Sigma} are sampled (\eqn{\omega} is omitted)}
}
\value{
\link{matrix} BPS posterior samples
}
\description{
Compute the BPS posterior samples given a set of stacking weights
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/code.R
\name{conv_opt}
\alias{conv_opt}
\title{Solver for Bayesian Predictive Stacking of Predictive densities convex optimization problem}
\usage{
conv_opt(scores)
}
\arguments{
\item{scores}{\link{matrix} \eqn{N \times K} of expected predictive density evaluations for the K models considered}
}
\value{
W \link{matrix} of Bayesian Predictive Stacking weights for the K models considered
}
\description{
Solver for Bayesian Predictive Stacking of Predictive densities convex optimization problem
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{BPS_PseudoBMA}
\alias{BPS_PseudoBMA}
\title{Combine subset models wiht Pseudo-BMA}
\usage{
BPS_PseudoBMA(fit_list)
}
\arguments{
\item{fit_list}{\link{list} K fitted model outputs composed by two elements each: first named \eqn{epd}, second named \eqn{W}}
}
\value{
\link{matrix} posterior predictive density evaluations (each columns represent a different model)
}
\description{
Combine subset models wiht Pseudo-BMA
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{BPS_combine}
\alias{BPS_combine}
\title{Combine subset models wiht BPS}
\usage{
BPS_combine(fit_list, K, rp)
}
\arguments{
\item{fit_list}{\link{list} K fitted model outputs composed by two elements each: first named \eqn{epd}, second named \eqn{W}}

\item{K}{\link{integer} number of folds}

\item{rp}{\link{double} percentage of observations to take into account for optimization (\code{default=1})}
}
\value{
\link{matrix} posterior predictive density evaluations (each columns represent a different model)
}
\description{
Combine subset models wiht BPS
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{dens_loocv_MvT}
\alias{dens_loocv_MvT}
\title{Compute the LOOCV of the density evaluations for fixed values of the hyperparameters}
\usage{
dens_loocv_MvT(data, priors, coords, hyperpar)
}
\arguments{
\item{data}{\link{list} two elements: first named \eqn{Y}, second named \eqn{X}}

\item{priors}{\link{list} priors: named \eqn{\mu_B},\eqn{V_r},\eqn{\Psi},\eqn{\nu}}

\item{coords}{\link{matrix} sample coordinates for X and Y}

\item{hyperpar}{\link{list} two elemets: first named \eqn{\alpha}, second named \eqn{\phi}}
}
\value{
\link{vector} posterior predictive density evaluations
}
\description{
Compute the LOOCV of the density evaluations for fixed values of the hyperparameters
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{models_dens_MvT}
\alias{models_dens_MvT}
\title{Return the CV predictive density evaluations for all the model combinations}
\usage{
models_dens_MvT(data, priors, coords, hyperpar, useKCV, K)
}
\arguments{
\item{data}{\link{list} two elements: first named \eqn{Y}, second named \eqn{X}}

\item{priors}{\link{list} priors: named \eqn{\mu_B},\eqn{V_r},\eqn{\Psi},\eqn{\nu}}

\item{coords}{\link{matrix} sample coordinates for X and Y}

\item{hyperpar}{\link{list} two elemets: first named \eqn{\alpha}, second named \eqn{\phi}}

\item{useKCV}{if \code{TRUE} K-fold cross validation is used instead of LOOCV (no \code{default})}

\item{K}{\link{integer} number of folds}
}
\value{
\link{matrix} posterior predictive density evaluations (each columns represent a different model)
}
\description{
Return the CV predictive density evaluations for all the model combinations
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{BPS_post_MvT}
\alias{BPS_post_MvT}
\title{Perform the BPS sampling from posterior and posterior predictive given a set of stacking weights}
\usage{
BPS_post_MvT(data, X_u, priors, coords, crd_u, hyperpar, W, R)
}
\arguments{
\item{data}{\link{list} two elements: first named \eqn{Y}, second named \eqn{X}}

\item{X_u}{\link{matrix} unobserved instances covariate matrix}

\item{priors}{\link{list} priors: named \eqn{\mu_B},\eqn{V_r},\eqn{\Psi},\eqn{\nu}}

\item{coords}{\link{matrix} sample coordinates for X and Y}

\item{crd_u}{\link{matrix} unboserved instances coordinates}

\item{hyperpar}{\link{list} two elemets: first named \eqn{\alpha}, second named \eqn{\phi}}

\item{W}{\link{matrix} set of stacking weights}

\item{R}{\link{integer} number of desired samples}
}
\value{
\link{list} BPS posterior predictive samples
}
\description{
Perform the BPS sampling from posterior and posterior predictive given a set of stacking weights
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bayesMvLMconjugate.R
\name{bayesMvLMconjugate}
\alias{bayesMvLMconjugate}
\title{Gibbs sampler for Conjugate Bayesian Multivariate Linear Models}
\usage{
bayesMvLMconjugate(Y, X, mu_B, V_B, nu, Psi, n_iter = 1000, burn_in = 500)
}
\arguments{
\item{Y}{\link{matrix} \eqn{n \times q} of response variables}

\item{X}{\link{matrix} \eqn{n \times p} of predictors}

\item{mu_B}{\link{matrix} \eqn{p \times q} prior mean for \eqn{\beta}}

\item{V_B}{\link{matrix} \eqn{p \times p} prior row covariance for \eqn{\beta}}

\item{nu}{\link{double} prior parameter for \eqn{\Sigma}}

\item{Psi}{\link{matrix} prior parameter for \eqn{\Sigma}}

\item{n_iter}{\link{integer} iteration number for Gibbs sampler}

\item{burn_in}{\link{integer} number of burn-in iteration}
}
\value{
B_samples \link{array} of posterior sample for \eqn{\beta}

Sigma_samples \link{array} of posterior samples for \eqn{\Sigma}
}
\description{
Gibbs sampler for Conjugate Bayesian Multivariate Linear Models
}
\examples{
\donttest{
## Generate data
n <- 100
p <- 3
q <- 2
Y <- matrix(rnorm(n*q), nrow = n, ncol = q)
X <- matrix(rnorm(n*p), nrow = n, ncol = p)

## Prior parameters
mu_B <- matrix(0, p, q)
V_B <- diag(10, p)
nu <- 3
Psi <- diag(q)

## Samples from posteriors
n_iter <- 1000
burn_in <- 500
set.seed(1234)
samples <- spBPS::bayesMvLMconjugate(Y, X, mu_B, V_B, nu, Psi, n_iter, burn_in)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{post_draws_MvT}
\alias{post_draws_MvT}
\title{Sample R draws from the posterior distributions}
\usage{
post_draws_MvT(poster, R, par, p)
}
\arguments{
\item{poster}{\link{list} output from \code{fit_cpp} function}

\item{R}{\link{integer} number of posterior samples}

\item{par}{if \code{TRUE} only \eqn{\beta} and \eqn{\Sigma} are sampled (\eqn{\omega} is omitted)}

\item{p}{\link{integer} if \code{par = TRUE}, it specifies the column number of \eqn{X}}
}
\value{
\link{list} posterior samples
}
\description{
Sample R draws from the posterior distributions
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bayesMvLMconjugate.R
\name{pred_bayesMvLMconjugate}
\alias{pred_bayesMvLMconjugate}
\title{Predictive sampler for Conjugate Bayesian Multivariate Linear Models}
\usage{
pred_bayesMvLMconjugate(X_new, B_samples, Sigma_samples)
}
\arguments{
\item{X_new}{\link{matrix} \eqn{n_new \times p} of predictors for new data points}

\item{B_samples}{\link{array} of posterior sample for \eqn{\beta}}

\item{Sigma_samples}{\link{array} of posterior samples for \eqn{\Sigma}}
}
\value{
Y_pred \link{matrix} of posterior mean for response matrix Y predictions

Y_pred_samples \link{array} of posterior predictive sample for response matrix Y
}
\description{
Predictive sampler for Conjugate Bayesian Multivariate Linear Models
}
\examples{
\donttest{
## Generate data
n <- 100
p <- 3
q <- 2
Y <- matrix(rnorm(n*q), nrow = n, ncol = q)
X <- matrix(rnorm(n*p), nrow = n, ncol = p)

## Prior parameters
mu_B <- matrix(0, p, q)
V_B <- diag(10, p)
nu <- 3
Psi <- diag(q)

## Samples from posteriors
n_iter <- 1000
burn_in <- 500
set.seed(1234)
samples <- spBPS::bayesMvLMconjugate(Y, X, mu_B, V_B, nu, Psi, n_iter, burn_in)

## Extract posterior samples
B_samples <- samples$B_samples
Sigma_samples <- samples$Sigma_samples

## Samples from predictive posterior (based posterior samples)
m <- 50
X_new <- matrix(rnorm(m*p), nrow = m, ncol = p)
pred <- spBPS::pred_bayesMvLMconjugate(X_new, B_samples, Sigma_samples)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/predict_spBPS_new.R
\name{predict.spBPS}
\alias{predict.spBPS}
\title{Predict at new locations using a fitted spBPS model}
\usage{
\method{predict}{spBPS}(
  object,
  newdata,
  draws = 200L,
  pred_batch_size = 200L,
  n_cores = 1L,
  return_summary = FALSE,
  probs = c(0.025, 0.975),
  ...
)
}
\arguments{
\item{object}{Object of class \code{"spBPS"} returned by \code{spBPS()}.}

\item{newdata}{List with elements \code{X} (u x p covariate matrix) and \code{coords} (u x d coordinate matrix).}

\item{draws}{Integer. Number of posterior predictive draws. Default 200.}

\item{pred_batch_size}{Integer. Number of prediction sites processed per batch. Default 200. Smaller values use less RAM; larger values may be faster.}

\item{n_cores}{Integer. Number of OMP threads. Default 1.}

\item{return_summary}{Logical. If \code{FALSE} (default) return full draw arrays. If \code{TRUE} return summary statistics.}

\item{probs}{Numeric vector of length 2. Quantile probabilities for credible intervals when \code{return_summary = TRUE}. Default \code{c(0.025, 0.975)}.}

\item{...}{Currently unused.}
}
\value{
When \code{return_summary = FALSE}, a list with arrays \code{Wu},
\code{Yu}, \code{MY} of dimension u x q x R.
When \code{return_summary = TRUE}, a list with u x q matrices
\code{mu_Yu}, \code{sd_Yu}, \code{q_lo}, \code{q_hi}, \code{mu_Wu}.
}
\description{
Generates posterior predictive samples at new spatial locations using a
fitted \code{spBPS} object. Memory usage is controlled via
\code{pred_batch_size}
}
\examples{
\donttest{
n <- 1000
p <- 2
q <- 1

Y <- matrix(rnorm(n * q), ncol = q)
X <- matrix(rnorm(n * p), ncol = p)
coords <- matrix(runif(n * 2), ncol = 2)

data    <- list(Y = Y, X = X)
priors  <- list(mu_B = matrix(0, nrow = p, ncol = q),
                V_r  = diag(10, p),
                Psi  = diag(1, q),
                nu   = 3)
hyperpar <- list(alpha = 0.5, phi = 1)

res <- spBPS(data, priors, coords, hyperpar, subset_size = 200)

u <- 500
p <- 2
q <- 1

X_u <- matrix(rnorm(u * p), ncol = p)
crd_u <- matrix(runif(u * 2), ncol = 2)

# Predictive sampling
pred <- predict(res, newdata=list(X=X_u, coords=crd_u),
                draws=200L, pred_batch_size=500L)

# Summary statistics only (memory-efficient for large u)
pred_sum <- predict(res, newdata=list(X=X_u, coords=crd_u),
                    draws=500L, return_summary=TRUE,
                    probs=c(0.025, 0.975))
}


}
\seealso{
\code{\link{spBPS}}
}
